from threading import Thread
from time import sleep
from typing import Any  # For DisposableObservable

from baseTest import BaseTestCase
from jstreams import (
    BehaviorSubject,
    Flowable,
    Observable,  # Added for DisposableObservable in merge tests
    PublishSubject,
    ReplaySubject,
    Single,
    RX,
    Timestamped,
)
from jstreams.eventing import event, events, managed_events, on_event
from jstreams.rx import (
    _EmptyObservable,  # For RX.merge tests
    BackpressureException,
    BackpressureMismatchException,
    BackpressureStrategy,
    SingleValueSubject,
    Subscribable,
)
from jstreams.utils import Value


class TestException(Exception):
    pass


class TestRx(BaseTestCase):
    def __init__(self, methodName="runTest"):
        super().__init__(methodName)

    def test_single(self) -> None:
        val = Value(None)
        Single("test").subscribe(val.set)
        self.assertEqual(val.get(), "test")

    def test_flowable(self) -> None:
        val = []
        init = ["test1", "test2"]
        Flowable(init).subscribe(val.append)
        self.assertListEqual(init, val)

    def test_behavior_subject(self) -> None:
        subject = BehaviorSubject("1")
        val = []
        sub = subject.subscribe(val.append)
        self.assertListEqual(
            val,
            ["1"],
            "BehaviorSubject should push the latest value on subscription",
        )
        subject.on_next("2")
        self.assertListEqual(
            val,
            ["1", "2"],
            "BehaviorSubject should push the latest value after subscription",
        )
        subject.on_next("3")
        self.assertListEqual(
            val,
            ["1", "2", "3"],
            "BehaviorSubject should push the latest value after subscription",
        )
        subject.pause(sub)
        subject.on_next("4")
        self.assertListEqual(
            val,
            ["1", "2", "3"],
            "BehaviorSubject should not push the latest value when subscription is paused",
        )
        subject.resume_paused()
        subject.on_next("5")
        self.assertListEqual(
            val,
            ["1", "2", "3", "5"],
            "BehaviorSubject should push the latest value when subscription is resumed",
        )
        subject.dispose()

    def test_publish_subject(self) -> None:
        subject = PublishSubject(str)
        val = []
        subject.on_next("1")
        sub = subject.subscribe(val.append)
        self.assertListEqual(
            val,
            [],
            "PublishSubject should not push the latest value on subscription",
        )
        subject.on_next("2")
        self.assertListEqual(
            val,
            ["2"],
            "PublishSubject should push the latest value after subscription",
        )
        subject.on_next("3")
        self.assertListEqual(
            val,
            ["2", "3"],
            "PublishSubject should push the latest value after subscription",
        )
        subject.pause(sub)
        subject.on_next("4")
        self.assertListEqual(
            val,
            ["2", "3"],
            "PublishSubject should not push the latest value when subscription is paused",
        )
        subject.resume_paused()
        subject.on_next("5")
        self.assertListEqual(
            val,
            ["2", "3", "5"],
            "PublishSubject should push the latest value when subscription is resumed",
        )
        subject.dispose()

    def test_replay_subject(self) -> None:
        subject = ReplaySubject(["A", "B", "C"])
        val = []
        val2 = []
        subject.subscribe(val.append)
        self.assertListEqual(val, ["A", "B", "C"])
        subject.on_next("1")
        self.assertListEqual(val, ["A", "B", "C", "1"])
        subject.subscribe(val2.append)
        self.assertListEqual(val2, ["A", "B", "C", "1"])
        subject.dispose()

    def test_replay_subject_map(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3"])
        val = []
        subject.pipe(RX.map(str.upper)).subscribe(val.append)
        self.assertListEqual(val, ["A1", "A2", "A3"])
        subject.on_next("a4")
        self.assertListEqual(val, ["A1", "A2", "A3", "A4"])
        subject.dispose()

    def test_replay_subject_filter(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3", "b", "c", "a4"])
        val = []
        subject.pipe(RX.filter(lambda s: s.startswith("a"))).subscribe(val.append)
        self.assertListEqual(val, ["a1", "a2", "a3", "a4"])
        subject.on_next("a5")
        self.assertListEqual(val, ["a1", "a2", "a3", "a4", "a5"])
        subject.on_next("b")
        self.assertListEqual(val, ["a1", "a2", "a3", "a4", "a5"])
        subject.dispose()

    def test_replay_subject_map_and_filter(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3"])
        val = []
        pipe1 = subject.pipe(RX.map(str.upper), RX.filter(lambda s: s.endswith("3")))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, ["A3"])
        subject.on_next("a4")
        self.assertListEqual(val, ["A3"])
        subject.dispose()

    def test_replay_subject_map_and_filter_multiple(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3"])
        val = []
        pipe1 = subject.pipe(
            RX.map(str.upper),
            RX.filter(lambda s: s.endswith("3")),
            RX.map(lambda s: s + "Test"),
        )
        pipe1.subscribe(val.append)
        self.assertListEqual(val, ["A3Test"])
        subject.on_next("a4")
        self.assertListEqual(val, ["A3Test"])
        subject.dispose()

    def test_replay_subject_filter_and_reduce(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(
            RX.filter(lambda nr: nr <= 10), RX.reduce(lambda a, b: max(a, b))
        )
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [1, 7])
        subject.on_next(9)
        self.assertListEqual(val, [1, 7, 9])
        subject.dispose()

    def test_replay_subject_with_take(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(RX.take(int, 3))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [1, 7, 20])
        subject.on_next(9)
        self.assertListEqual(val, [1, 7, 20])
        subject.dispose()

    def test_replay_subject_with_takeWhile(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(RX.take_while(lambda v: v < 10))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [1, 7])
        subject.on_next(9)
        self.assertListEqual(val, [1, 7])
        subject.dispose()

    def test_replay_subject_with_takeUntil(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(RX.take_until(lambda v: v > 10))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [1, 7])
        subject.on_next(9)
        self.assertListEqual(val, [1, 7])
        subject.dispose()

    def test_replay_subject_with_drop(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(RX.drop(int, 3))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [5, 100, 40])
        subject.on_next(9)
        self.assertListEqual(val, [5, 100, 40, 9])
        subject.dispose()

    def test_replay_subject_with_dropWhile(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(RX.drop_while(lambda v: v < 100))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [100, 40])
        subject.on_next(9)
        self.assertListEqual(val, [100, 40, 9])
        subject.dispose()

    def test_replay_subject_with_dropUntil(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        val = []
        pipe1 = subject.pipe(RX.drop_until(lambda v: v > 20))
        pipe1.subscribe(val.append)
        self.assertListEqual(val, [100, 40])
        subject.on_next(9)
        self.assertListEqual(val, [100, 40, 9])
        subject.dispose()

    def test_replay_subject_with_pipe_chaining(self) -> None:
        subject = ReplaySubject(range(1, 100))
        val = []
        val2 = []
        chainedPipe = (
            subject.pipe(RX.take_until(lambda e: e > 20))
            .pipe(RX.filter(lambda e: e % 2 == 0))
            .pipe(RX.take_while(lambda e: e < 10))
        )
        chainedPipe.subscribe(val.append)
        chainedPipe.subscribe(val2.append)
        chainedPipe.dispose()
        self.assertListEqual(val, [2, 4, 6, 8])
        self.assertListEqual(val2, [2, 4, 6, 8])

    def test_event_cancelling_subs(self) -> None:
        self.__test_event(True)

    def test_event_cancelling_events(self) -> None:
        self.__test_event(False)

    def __test_event(self, subs: bool) -> None:
        disposed_val = Value(False)
        disposed_valsub = Value(False)

        vals = []
        valsub = []

        subpipe = (
            event(int)
            .pipe(RX.map(lambda i: str(i)))
            .subscribe(vals.append, on_dispose=lambda: disposed_val.set(True))
        )
        subval = event(int).subscribe(valsub.append, lambda: disposed_valsub.set(True))

        self.assertIsNotNone(subpipe)
        self.assertIsNotNone(subval)

        event(int).publish(1)
        event(int).publish(2)
        sleep(1)
        if subs:
            subpipe.cancel()
            subval.cancel()
            subpipe.dispose()
        else:
            events().clear_event(int)

        event(int).publish(3)
        sleep(1)
        self.assertEqual(event(int).latest(), 3)

        self.assertListEqual(vals, ["1", "2"])
        self.assertListEqual(valsub, [1, 2])
        if subs:
            self.assertTrue(disposed_val.get())
            self.assertTrue(disposed_valsub.get())

    def test_managed_events(self) -> None:
        @managed_events()
        class TestManagedEvents:
            def __init__(self):
                self.int_value = None
                self.str_value = None

            @on_event(int)
            def on_int_event(self, value: int) -> None:
                self.int_value = value

            @on_event(str)
            def on_str_event(self, value: str) -> None:
                self.str_value = value

        test = TestManagedEvents()
        self.assertIsNone(test.int_value)
        self.assertIsNone(test.str_value)
        event(int).publish(10)
        event(str).publish("test")
        sleep(1)
        self.assertEqual(test.int_value, 10)
        self.assertEqual(test.str_value, "test")
        del test

    def test_publish_if(self) -> None:
        val = Value(None)
        event(int).subscribe(val.set)
        event(int).publish_if(1, lambda _: True)
        sleep(1)
        self.assertEqual(val.get(), 1)

        val.set(None)
        event(int).publish_if(2, lambda _: False)
        sleep(1)
        self.assertEqual(val.get(), None)

    def test_has_events(self) -> None:
        self.assertFalse(events().has_event(str))
        event(str).publish("test")
        self.assertTrue(events().has_event(str))
        events().clear_event(str)
        self.assertFalse(events().has_event(str))

    def test_subscribe_once(self) -> None:
        elements = []
        event(str).subscribe_once(elements.append)
        event(str).publish("test")
        event(str).publish("test2")
        sleep(1)
        self.assertListEqual(elements, ["test"])

    def test_distinct_until_changed(self) -> None:
        elements = []
        event(str).pipe(RX.distinct_until_changed(str)).subscribe(elements.append)
        event(str).publish("test")
        event(str).publish("test")
        event(str).publish("test2")
        sleep(1)
        self.assertListEqual(elements, ["test", "test2"])

    def test_distinct_until_changed_with_key(self) -> None:
        elements = []
        event(str).pipe(RX.distinct_until_changed(str, lambda s: s[0])).subscribe(
            elements.append
        )
        event(str).publish("test")
        event(str).publish("test2")
        event(str).publish("best")
        sleep(1)
        self.assertListEqual(elements, ["test", "best"])
        event(str).publish("test3")
        event(str).publish("test4")
        sleep(1)
        self.assertListEqual(elements, ["test", "best", "test3"])

    def test_tap(self) -> None:
        elements = []
        event(str).pipe(RX.tap(elements.append)).subscribe(lambda _: None)
        event(str).publish("test")
        event(str).publish("test2")
        sleep(1)
        self.assertListEqual(elements, ["test", "test2"])

    def test_ignore_all(self) -> None:
        elements = []
        event(str).pipe(RX.ignore_all()).subscribe(elements.append)
        event(str).publish("test")
        event(str).publish("test2")
        sleep(1)
        self.assertListEqual(elements, [])

    def test_ignore(self) -> None:
        elements = []
        event(str).pipe(RX.ignore(lambda _: True)).subscribe(elements.append)
        event(str).publish("test")
        event(str).publish("test2")
        sleep(1)
        self.assertListEqual(elements, [])

        elements = []
        event(str, "1").pipe(RX.ignore(lambda _: False)).subscribe(elements.append)
        event(str, "1").publish("test")
        event(str, "1").publish("test2")
        sleep(1)
        self.assertListEqual(elements, ["test", "test2"])

        elements = []
        event(str, "2").pipe(RX.ignore(lambda e: e.startswith("t"))).subscribe(
            elements.append
        )
        event(str, "2").publish("test")
        event(str, "2").publish("best")
        sleep(1)
        self.assertListEqual(elements, ["best"])

    def test_rx_empty(self) -> None:
        on_next_called = Value(False)
        on_completed_called = Value(False)
        on_error_called = Value(False)

        RX.empty().subscribe(
            lambda _: on_next_called.set(True),
            lambda _: on_error_called.set(True),
            lambda _: on_completed_called.set(True),
        )
        self.assertFalse(on_next_called.get())
        self.assertTrue(on_completed_called.get())
        self.assertFalse(on_error_called.get())

    def test_rx_never(self) -> None:
        on_next_called = Value(False)
        on_completed_called = Value(False)
        on_error_called = Value(False)

        RX.never().subscribe(
            lambda _: on_next_called.set(True),
            lambda _: on_error_called.set(True),
            lambda _: on_completed_called.set(True),
        )
        sleep(0.01)  # Give a small time for any potential async emission
        self.assertFalse(on_next_called.get())
        self.assertFalse(on_completed_called.get())
        self.assertFalse(on_error_called.get())

    def test_rx_throw(self) -> None:
        on_next_called = Value(False)
        on_completed_called = Value(False)
        error_val = Value(None)

        class TestException(Exception):
            pass

        test_exception = TestException("test error")

        RX.throw(test_exception).subscribe(
            lambda _: on_next_called.set(True),
            error_val.set,
            lambda _: on_completed_called.set(True),
        )
        self.assertFalse(on_next_called.get())
        self.assertFalse(on_completed_called.get())
        self.assertIsInstance(error_val.get(), TestException)
        self.assertEqual(str(error_val.get()), "test error")

        # Test with factory
        on_next_called.set(False)
        on_completed_called.set(False)
        error_val.set(None)
        RX.throw(lambda: TestException("factory error")).subscribe(
            lambda _: on_next_called.set(True),
            error_val.set,
            lambda _: on_completed_called.set(True),
        )
        self.assertFalse(on_next_called.get())
        self.assertFalse(on_completed_called.get())
        self.assertIsInstance(error_val.get(), TestException)
        self.assertEqual(str(error_val.get()), "factory error")

    def test_rx_range(self) -> None:
        results = []
        RX.range(1, 5).subscribe(results.append)
        self.assertListEqual(results, [1, 2, 3, 4, 5])

        results = []
        RX.range(0, 0).subscribe(results.append)  # Should behave like empty
        self.assertListEqual(results, [])

        with self.assertRaises(ValueError):
            RX.range(1, -1)

    def test_rx_defer(self) -> None:
        factory_called_count = Value(0)

        def observable_factory():
            factory_called_count.set(factory_called_count.get() + 1)
            return Flowable([1, 2])

        deferred_obs = RX.defer(observable_factory)

        results1 = []
        deferred_obs.subscribe(results1.append)
        self.assertEqual(factory_called_count.get(), 1)
        self.assertListEqual(results1, [1, 2])

        results2 = []
        deferred_obs.subscribe(results2.append)
        self.assertEqual(factory_called_count.get(), 2)
        self.assertListEqual(results2, [1, 2])

        # Test defer with factory error
        error_val = Value(None)

        def error_factory():
            raise TestException("factory failed")

        RX.defer(error_factory).subscribe(lambda _: None, error_val.set)
        self.assertIsInstance(error_val.get(), TestException)
        self.assertEqual(str(error_val.get()), "factory failed")

    def test_rx_map_to(self) -> None:
        results = []
        Flowable([1, 2, 3]).pipe(RX.map_to("A")).subscribe(results.append)
        self.assertListEqual(results, ["A", "A", "A"])

    def test_rx_scan(self) -> None:
        results = []
        Flowable([1, 2, 3, 4]).pipe(RX.scan(lambda acc, val: acc + val, 0)).subscribe(
            results.append
        )
        self.assertListEqual(results, [1, 3, 6, 10])

        results = []
        Flowable([1, 2, 3]).pipe(RX.scan(lambda acc, val: acc + [val], [])).subscribe(
            results.append
        )
        self.assertListEqual(results, [[1], [1, 2], [1, 2, 3]])

    def test_rx_distinct(self) -> None:
        results = []
        Flowable([1, 2, 2, 3, 1, 4, 4]).pipe(RX.distinct(int)).subscribe(results.append)
        self.assertListEqual(results, [1, 2, 3, 4])

        results = []
        data = [{"id": 1, "val": "a"}, {"id": 2, "val": "b"}, {"id": 1, "val": "c"}]
        Flowable(data).pipe(
            RX.distinct(dict[str, str], key_selector=lambda x: x["id"])
        ).subscribe(results.append)
        self.assertListEqual(results, [{"id": 1, "val": "a"}, {"id": 2, "val": "b"}])

    def test_rx_timestamp(self) -> None:
        results = []
        start_time = Value(None)
        Flowable(["a", "b"]).pipe(RX.timestamp(str)).subscribe(
            lambda ts_val: (
                start_time.set(ts_val.timestamp) if start_time.get() is None else None,
                results.append(ts_val),
            )
        )
        self.assertEqual(len(results), 2)
        self.assertIsInstance(results[0], Timestamped)
        self.assertEqual(results[0].value, "a")
        self.assertGreaterEqual(results[0].timestamp, start_time.get())
        self.assertIsInstance(results[1], Timestamped)
        self.assertEqual(results[1].value, "b")
        self.assertGreaterEqual(results[1].timestamp, results[0].timestamp)

    def test_rx_element_at(self) -> None:
        results = []
        Flowable(["a", "b", "c"]).pipe(RX.element_at(str, 1)).subscribe(results.append)
        self.assertListEqual(results, ["b"])

        results = []
        Flowable(["a", "b", "c"]).pipe(RX.element_at(str, 0)).subscribe(results.append)
        self.assertListEqual(results, ["a"])

        results = []
        Flowable(["a", "b", "c"]).pipe(RX.element_at(str, 2)).subscribe(results.append)
        self.assertListEqual(results, ["c"])

        results = []  # out of bounds
        Flowable(["a", "b", "c"]).pipe(RX.element_at(str, 3)).subscribe(results.append)
        self.assertListEqual(results, [])

        with self.assertRaises(ValueError):
            RX.element_at(str, -1)

    def test_buffer_emits_none_then_list(self) -> None:
        pipe = RX.buffer(int, 0.02)
        pipe.init()  # Important for stateful operators like buffer
        self.assertIsNone(pipe.transform(1))  # type: ignore
        self.assertIsNone(pipe.transform(2))  # type: ignore
        sleep(0.03)
        self.assertListEqual(pipe.transform(3), [1, 2, 3])  # type: ignore
        self.assertIsNone(pipe.transform(4))  # type: ignore

    def test_buffer_count_emits_none_then_list(self) -> None:
        pipe = RX.buffer_count(int, 3)
        pipe.init()
        self.assertIsNone(pipe.transform(1))  # type: ignore
        self.assertIsNone(pipe.transform(2))  # type: ignore
        self.assertListEqual(pipe.transform(3), [1, 2, 3])  # type: ignore
        self.assertIsNone(pipe.transform(4))  # type: ignore

    def test_buffer_count_emits_none_then_list_subject(self) -> None:
        subject = SingleValueSubject(1)
        value_list = Value([])
        subject.pipe(RX.buffer_count(int, 3)).subscribe(value_list.set)
        subject.on_next(2)
        subject.on_next(3)
        subject.on_next(4)
        subject.on_next(5)
        self.assertListEqual(value_list.get(), [1, 2, 3])
        subject.on_next(6)
        self.assertListEqual(value_list.get(), [4, 5, 6])
        subject.dispose()

    def test_backpressure_drop(self) -> None:
        subject = SingleValueSubject("test")
        values = []
        errors = []

        def handler(val: str) -> None:
            values.append(val)
            sleep(1)

        def error(val: Exception) -> None:
            errors.append(val)

        subject.subscribe(
            handler,
            on_error=error,
            backpressure=BackpressureStrategy.DROP,
            asynchronous=True,
        )
        subject.on_next("test1")
        subject.on_next("test2")
        sleep(1.5)

        self.assertEqual(values, ["test"])
        self.assertEqual(errors, [])

    def test_backpressure_error(self) -> None:
        subject = SingleValueSubject("test")
        values = []
        errors = []

        def handler(val: str) -> None:
            values.append(val)
            sleep(1)

        def error(val: Exception) -> None:
            errors.append(val)

        subject.subscribe(
            handler,
            on_error=error,
            backpressure=BackpressureStrategy.ERROR,
            asynchronous=True,
        )
        subject.on_next("test1")
        subject.on_next("test2")
        sleep(1.5)

        self.assertEqual(values, ["test"])
        self.assertEqual(
            errors,
            [
                BackpressureException("Missed value"),
                BackpressureException("Missed value"),
            ],
        )

    def test_invalid_backpressure(self) -> None:
        subject = SingleValueSubject("test")
        self.assertRaises(
            BackpressureMismatchException,
            lambda: subject.subscribe(
                lambda _: None,
                asynchronous=False,
                backpressure=BackpressureStrategy.DROP,
            ),
        )


class TestRxMerge(BaseTestCase):
    def test_merge_basic(self) -> None:
        s1 = Flowable([1, 2, 3])
        s2 = Flowable(["a", "b", "c"])
        results = []
        completed = Value(False)

        RX.merge(s1, s2).subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        s1.on_completed(None)
        s2.on_completed(None)

        self.assertCountEqual(results, [1, 2, 3, "a", "b", "c"])
        self.assertTrue(completed.get())

    def test_merge_with_empty_source(self) -> None:
        s1 = Flowable([1, 2])
        s2 = RX.empty()
        s3 = Flowable(["a", "b"])
        results = []
        completed = Value(False)

        RX.merge(s1, s2, s3).subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        s1.on_completed(None)
        s3.on_completed(None)

        self.assertCountEqual(results, [1, 2, "a", "b"])
        self.assertTrue(completed.get())

    def test_merge_all_empty_sources(self) -> None:
        s1 = RX.empty()
        s2 = RX.empty()
        results = []
        completed = Value(False)

        RX.merge(s1, s2).subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        self.assertEqual(len(results), 0)
        self.assertTrue(completed.get())

    def test_merge_no_sources(self) -> None:
        merged = RX.merge()
        self.assertIsInstance(
            merged,
            _EmptyObservable,
            "Merging no sources should return an empty observable",
        )
        results = []
        completed = Value(False)
        merged.subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        self.assertEqual(len(results), 0)
        self.assertTrue(completed.get())

    def test_merge_single_source(self) -> None:
        s1 = Flowable([1, 2, 3])
        merged = RX.merge(s1)
        self.assertIs(
            merged, s1, "Merging a single source should return the source itself"
        )

        results = []
        completed = Value(False)
        merged.subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        s1.on_completed(None)
        self.assertEqual(results, [1, 2, 3])
        self.assertTrue(completed.get())

    def test_merge_completes_after_all_sources_complete(self) -> None:
        s1_ps = PublishSubject(int)
        s2_ps = PublishSubject(int)

        s1 = s1_ps.pipe(RX.map(lambda x: f"s1:{x}"))
        s2 = s2_ps.pipe(RX.map(lambda x: f"s2:{x}"))

        results = []
        completed = Value(False)

        sub = RX.merge(s1, s2).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        def s1_emitter():
            for i in range(2):  # s1:0, s1:1
                sleep(0.02)
                s1_ps.on_next(i)
            s1_ps.dispose()

        def s2_emitter():
            for i in range(3):  # s2:0, s2:1, s2:2
                sleep(0.03)
                s2_ps.on_next(i)
            s2_ps.dispose()

        t1 = Thread(target=s1_emitter)
        t2 = Thread(target=s2_emitter)
        t1.start()
        t2.start()

        t1.join(timeout=0.2)
        t2.join(timeout=0.2)
        sleep(1)  # Ensure completion callback has time to fire

        # self.assertTrue(completed.get(), "Should complete after all sources complete")
        self.assertCountEqual(results, ["s1:0", "s1:1", "s2:0", "s2:1", "s2:2"])
        sub.cancel()

    def test_merge_errors_if_one_source_errors(self) -> None:
        s1_ps = PublishSubject(int)
        s2_ps = PublishSubject(int)

        s1 = s1_ps.pipe(RX.take(int, 5), RX.map(lambda x: f"s1:{x}"))
        s2_error = s2_ps.pipe(
            RX.map(lambda x: {"val": x}),
            RX.map(
                lambda x: 1 / 0 if x["val"] == 1 else x
            ),  # Error on second item (val=1)
        )

        results = []
        error_received = Value(None)
        completed = Value(False)

        sub = RX.merge(s1, s2_error).subscribe(
            on_next=results.append,
            on_error=error_received.set,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        def s1_emitter():
            for i in range(5):
                sleep(0.01)
                if error_received.get():
                    break  # Stop if error occurred
                s1_ps.on_next(i)
            s1_ps.on_completed(None)

        def s2_emitter():
            s2_ps.on_next(0)  # {"val": 0}
            sleep(0.005)
            s2_ps.on_next(1)  # This will cause error
            s2_ps.on_completed(None)

        t1 = Thread(target=s1_emitter)
        t2 = Thread(target=s2_emitter)
        t1.start()
        t2.start()

        t1.join(timeout=0.2)
        t2.join(timeout=0.2)
        sleep(0.05)

        self.assertIsNotNone(error_received.get())
        self.assertIsInstance(error_received.get(), ZeroDivisionError)
        self.assertFalse(completed.get(), "Should not complete if an error occurs")
        self.assertTrue(any(r == {"val": 0} for r in results))  # First item from s2
        sub.cancel()

    def test_merge_cancellation_disposes_inner_subscriptions(self) -> None:
        s1_dispose_called = Value(False)
        s2_dispose_called = Value(False)

        # Use PublishSubject to simulate long-running/infinite streams
        s1_ps = PublishSubject(int)
        s2_ps = PublishSubject(int)

        s1_source = s1_ps.pipe(RX.map(lambda x: f"s1:{x}"))
        s2_source = s2_ps.pipe(RX.map(lambda x: f"s2:{x}"))

        class DisposableObservable(Observable[Any]):  # Inherit from Observable
            def __init__(self, source: Subscribable[Any], dispose_flag: Value[bool]):
                super().__init__()
                self._source_subscribable = source
                self._dispose_flag = dispose_flag

            def subscribe(
                self,
                on_next=None,
                on_error=None,
                on_completed=None,
                on_dispose=None,
                asynchronous=False,
                backpressure=None,
            ):
                original_user_on_dispose = on_dispose

                def custom_dispose():
                    self._dispose_flag.set(True)
                    if original_user_on_dispose:
                        original_user_on_dispose()

                return self._source_subscribable.subscribe(
                    on_next,
                    on_error,
                    on_completed,
                    custom_dispose,
                    asynchronous,
                    backpressure,
                )

        disposable_s1 = DisposableObservable(s1_source, s1_dispose_called)
        disposable_s2 = DisposableObservable(s2_source, s2_dispose_called)

        merged_sub = RX.merge(disposable_s1, disposable_s2).subscribe(asynchronous=True)

        s1_ps.on_next(1)  # Emit something to ensure subscriptions are active
        sleep(0.05)
        merged_sub.cancel()
        sleep(0.05)

        self.assertTrue(s1_dispose_called.get(), "s1 should have been disposed")
        self.assertTrue(s2_dispose_called.get(), "s2 should have been disposed")

        s1_ps.on_completed(None)  # Clean up subjects
        s2_ps.on_completed(None)


class TestRxZip(BaseTestCase):
    def test_zip_basic_tuple(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(str)
        results = []
        completed = Value(False)
        errored = Value(False)

        sub = RX.zip(tuple[int, str], s1, s2).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            on_error=lambda _: errored.set(True),
            asynchronous=True,
        )

        s1.on_next(1)
        s1.on_next(2)  # s1 has 1, 2 queued
        s2.on_next("a")  # s2 has "a" queued. Emits (1, "a"). s1 has 2 left.
        sleep(0.05)
        self.assertEqual(results, [(1, "a")])

        s2.on_next("b")  # s2 has "b" queued. Emits (2, "b"). s1 is empty.
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "b")])

        s1.on_next(3)  # s1 has 3. s2 is empty. No emission.
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "b")])

        s1.on_completed(None)  # s1 completes. Still has 3 in queue.
        s2.on_next(
            "c"
        )  # s2 has "c". Emits (3, "c"). s1 queue empty. s1 completed. Zip completes.
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "b"), (3, "c")])
        self.assertTrue(completed.get())
        self.assertFalse(errored.get())
        sub.cancel()

    def test_zip_with_zipper_function(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(int)
        results = []

        def zipper_fn(v1: int, v2: int) -> str:
            return f"{v1}*{v2}={v1 * v2}"

        sub = RX.zip(str, s1, s2, zipper=zipper_fn).subscribe(
            on_next=results.append, asynchronous=True
        )

        s1.on_next(2)
        s2.on_next(10)  # Emits "2*10=20"
        sleep(0.05)
        self.assertEqual(results, ["2*10=20"])

        s1.on_next(3)
        s1.on_next(4)
        s2.on_next(5)  # Emits "3*5=15"
        sleep(0.05)
        self.assertEqual(results, ["2*10=20", "3*5=15"])
        sub.cancel()

    def test_zip_one_source_completes_early(self) -> None:
        s1 = Flowable([1, 2])  # Completes after 2
        s2 = PublishSubject(str)
        results = []
        completed = Value(False)

        sub = RX.zip(tuple[int, str], s1, s2).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        s2.on_next("a")  # (1, "a")
        sleep(0.05)
        self.assertEqual(results, [(1, "a")])

        s2.on_next("b")  # (2, "b"). s1 is now exhausted and completes. Zip completes.
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "b")])
        # self.assertTrue(completed.get())

        s2.on_next("c")  # Should have no effect as zip stream is completed
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "b")])
        sub.cancel()

    def test_zip_error_propagation(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(str)
        results = []
        error_received = Value(None)
        completed = Value(False)

        sub = RX.zip(tuple[int, str], s1, s2).subscribe(
            on_next=results.append,
            on_error=error_received.set,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        s1.on_next(1)
        s2.on_next("a")
        sleep(0.05)
        self.assertEqual(results, [(1, "a")])

        test_exception = TestException("Zip Source error")
        s1.on_error(test_exception)  # s1 errors
        sleep(0.05)

        self.assertIs(error_received.get(), test_exception)
        self.assertFalse(completed.get())
        s2.on_next("b")  # Should have no effect
        sleep(0.05)
        self.assertEqual(results, [(1, "a")])
        sub.cancel()

    def test_zip_no_sources(self) -> None:
        results = []
        completed = Value(False)
        sub = RX.zip(Any).subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        self.assertTrue(completed.get())
        self.assertEqual(len(results), 0)
        sub.cancel()

    def test_zip_single_source(self) -> None:
        s1 = PublishSubject(int)
        results = []
        completed = Value(False)

        sub = RX.zip(tuple[int], s1).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )
        s1.on_next(10)
        sleep(0.05)
        self.assertEqual(results, [(10,)])
        s1.on_next(20)
        sleep(0.05)
        self.assertEqual(results, [(10,), (20,)])
        s1.on_completed(None)
        sleep(0.05)
        self.assertTrue(completed.get())
        sub.cancel()

    def test_zip_completes_when_shortest_source_exhausted_after_zip(self) -> None:
        s1 = Flowable([1, 2])
        s2 = Flowable(["a", "b", "c"])
        results = []
        completed = Value(False)

        sub = RX.zip(tuple[int, str], s1, s2).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )
        sleep(0.1)

        self.assertEqual(results, [(1, "a"), (2, "b")])
        # self.assertTrue(completed.get())
        sub.cancel()

    def test_zip_with_empty_source(self) -> None:
        s1 = Flowable([1, 2, 3])
        s_empty = RX.empty()
        results = []
        completed = Value(False)

        sub = RX.zip(tuple[int, Any], s1, s_empty).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )
        sleep(0.1)
        self.assertEqual(len(results), 0)
        self.assertTrue(completed.get())
        sub.cancel()

    def test_merge_with_behavior_subject(self) -> None:
        bs = BehaviorSubject("initial_bs")
        fs = Flowable([1, 2])
        results = []
        completed = Value(False)

        sub = RX.merge(bs, fs).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )
        bs.on_next("bs_update1")
        bs.on_completed(None)
        sleep(0.1)
        # self.assertTrue(completed.get())
        self.assertIn("initial_bs", results)
        self.assertIn("bs_update1", results)
        self.assertIn(1, results)
        self.assertIn(2, results)
        self.assertEqual(len(results), 4)
        sub.cancel()

    def test_merge_with_publish_subject(self) -> None:
        ps = PublishSubject(str)
        fs = Flowable([10, 20])  # This will emit and complete quickly
        results = []
        completed = Value(False)

        sub = RX.merge(ps, fs).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        ps.on_next("ps_event1")
        # fs completes around here
        ps.on_next("ps_event2")
        ps.on_completed(None)  # Now ps completes

        sleep(0.1)  # Allow all events to process

        # self.assertTrue(completed.get())
        self.assertIn("ps_event1", results)
        self.assertIn("ps_event2", results)
        self.assertIn(10, results)
        self.assertIn(20, results)
        self.assertEqual(len(results), 4)
        sub.cancel()

    def test_merge_with_never(self) -> None:
        s1 = Flowable([1, 2])  # Completes
        s_never = RX.never()
        results = []
        completed = Value(False)
        errored = Value(False)

        sub = RX.merge(s1, s_never).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            on_error=lambda _: errored.set(True),
            asynchronous=True,
        )
        sleep(0.1)
        self.assertFalse(
            completed.get(), "Should not complete if one source never completes"
        )
        self.assertFalse(errored.get())
        self.assertCountEqual(results, [1, 2])
        sub.cancel()

    def test_merge_with_throw(self) -> None:
        s1 = Flowable([1, 2])
        s_throw = RX.throw(TestException("Merge Error Test"))
        results = []
        completed = Value(False)
        error_val = Value(None)

        sub = RX.merge(s1, s_throw).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            on_error=error_val.set,
            asynchronous=True,
        )
        sleep(0.1)
        self.assertFalse(completed.get())
        self.assertIsNotNone(error_val.get())
        self.assertIsInstance(error_val.get(), TestException)
        self.assertEqual(str(error_val.get()), "Merge Error Test")
        sub.cancel()

    def test_merge_handles_source_completing_before_subscribe_fully_processed_by_others(
        self,
    ) -> None:
        s1 = Flowable([1])  # Completes almost immediately
        s2_ps = PublishSubject(int)
        s2 = s2_ps.pipe(RX.map(lambda x: f"s2:{x}"))
        results = []
        completed = Value(False)

        sub = RX.merge(s1, s2).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        def s2_emitter():
            sleep(0.05)  # s1 has likely completed by now
            s2_ps.on_next(0)
            sleep(0.05)
            s2_ps.on_next(1)
            s2_ps.on_completed(None)

        Thread(target=s2_emitter).start()
        sleep(0.2)

        # self.assertTrue(completed.get())
        self.assertIn(1, results)
        self.assertIn("s2:0", results)
        self.assertIn("s2:1", results)
        self.assertEqual(len(results), 3)
        sub.cancel()

    # def test_merge_handles_source_erroring_before_subscribe_fully_processed_by_others(
    #     self,
    # ) -> None:
    #     s1 = RX.throw(TestException("Immediate Error"))
    #     s2_ps = PublishSubject(int)
    #     s2 = s2_ps.pipe(RX.map(lambda x: f"s2:{x}"))
    #     results = []
    #     completed = Value(False)
    #     error_val = Value(None)

    #     sub = RX.merge(s1, s2).subscribe(
    #         on_next=results.append,
    #         on_completed=lambda _: completed.set(True),
    #         on_error=error_val.set,
    #         asynchronous=True,
    #     )

    #     def s2_emitter():  # This emitter might not even get to run if s1 errors fast enough
    #         sleep(0.05)
    #         s2_ps.on_next(0)
    #         s2_ps.on_completed(None)

    #     Thread(target=s2_emitter).start()
    #     sleep(0.2)

    #     # self.assertFalse(completed.get())
    #     self.assertIsNotNone(error_val.get())
    #     self.assertIsInstance(error_val.get(), TestException)
    #     self.assertEqual(len(results), 0)  # s2 should not have emitted
    #     sub.cancel()
    #     s2_ps.on_completed(None)  # Clean up subject

    def test_throttle(self) -> None:
        subject = SingleValueSubject(1)
        vals = []
        subject.pipe(RX.throttle(int, 0.5)).subscribe(vals.append)
        subject.on_next(0)
        subject.on_next(1)
        subject.on_next(2)
        subject.on_next(4)
        subject.on_next(5)
        sleep(1)
        self.assertEqual(vals, [1])

    def test_throttle_tap(self) -> None:
        subject = SingleValueSubject(1)
        vals = []
        subject.pipe(RX.debounce(int, 0.5), RX.tap(vals.append)).subscribe()
        subject.on_next(0)
        subject.on_next(1)
        subject.on_next(2)
        subject.on_next(4)
        subject.on_next(5)
        sleep(1)
        self.assertEqual(vals, [5])


class TestRxCombineLatest(BaseTestCase):
    def test_combine_latest_basic_tuple(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(str)
        results = []
        completed = Value(False)

        sub = RX.combine_latest(
            tuple[int, str],
            s1,
            s2,
        ).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        s1.on_next(1)
        self.assertEqual(
            len(results), 0, "Should not emit until all sources emitted once"
        )

        s2.on_next("a")  # First emission from s2, now both have emitted
        sleep(0.05)
        self.assertEqual(results, [(1, "a")])

        s1.on_next(2)  # s2's latest is still "a"
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "a")])

        s2.on_next("b")  # s1's latest is 2
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "a"), (2, "b")])

        s1.on_completed(None)
        s2.on_next(
            "c"
        )  # s1 completed, but s2 can still trigger using s1's last value (2)
        sleep(0.05)
        self.assertEqual(results, [(1, "a"), (2, "a"), (2, "b"), (2, "c")])

        s2.on_completed(None)  # All sources completed
        sleep(0.05)
        self.assertTrue(completed.get())
        sub.cancel()

    def test_combine_latest_with_combiner_function(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(int)
        results = []

        def combiner(v1: int, v2: int) -> str:
            return f"{v1}+{v2}={v1 + v2}"

        sub = RX.combine_latest(str, s1, s2, combiner=combiner).subscribe(
            on_next=results.append, asynchronous=True
        )

        s1.on_next(1)
        s2.on_next(10)  # Emits "1+10=11"
        sleep(0.05)
        self.assertEqual(results, ["1+10=11"])

        s1.on_next(2)  # Emits "2+10=12"
        sleep(0.05)
        self.assertEqual(results, ["1+10=11", "2+10=12"])
        sub.cancel()

    def test_combine_latest_one_source_completes_before_emitting(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(str)
        results = []
        completed = Value(False)
        errored = Value(False)

        sub = RX.combine_latest(tuple[int, str], s1, s2).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            on_error=lambda _: errored.set(True),
            asynchronous=True,
        )

        s1.on_next(1)
        s2.on_completed(None)  # s2 completes before emitting anything
        sleep(0.05)

        self.assertTrue(
            completed.get(), "Should complete if one source completes before emitting"
        )
        self.assertFalse(errored.get())
        self.assertEqual(len(results), 0, "Should not emit any values")
        sub.cancel()

    def test_combine_latest_error_propagation(self) -> None:
        s1 = PublishSubject(int)
        s2 = PublishSubject(str)
        results = []
        error_received = Value(None)
        completed = Value(False)

        sub = RX.combine_latest(tuple[int, str], s1, s2).subscribe(
            on_next=results.append,
            on_error=error_received.set,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )

        s1.on_next(1)
        s2.on_next("a")
        sleep(0.05)
        self.assertEqual(results, [(1, "a")])

        test_exception = TestException("Source error")
        s1.on_error(test_exception)
        sleep(0.05)

        self.assertIs(error_received.get(), test_exception)
        self.assertFalse(completed.get())
        sub.cancel()

    def test_combine_latest_no_sources(self) -> None:
        results = []
        completed = Value(False)
        sub = RX.combine_latest(Any).subscribe(
            on_next=results.append, on_completed=lambda _: completed.set(True)
        )
        self.assertTrue(completed.get())
        self.assertEqual(len(results), 0)
        sub.cancel()  # Should be a no-op on an already completed empty sub

    def test_combine_latest_single_source(self) -> None:
        s1 = PublishSubject(int)
        results = []
        completed = Value(False)

        sub = RX.combine_latest(tuple[int], s1).subscribe(
            on_next=results.append,
            on_completed=lambda _: completed.set(True),
            asynchronous=True,
        )
        s1.on_next(10)
        sleep(0.05)
        self.assertEqual(results, [(10,)])
        s1.on_next(20)
        sleep(0.05)
        self.assertEqual(results, [(10,), (20,)])
        s1.on_completed(None)
        sleep(0.05)
        self.assertTrue(completed.get())
        sub.cancel()

    def test_combine_latest_single_source_with_combiner(self) -> None:
        s1 = PublishSubject(int)
        results = []
        sub = RX.combine_latest(str, s1, combiner=lambda x: f"val:{x}").subscribe(
            on_next=results.append, asynchronous=True
        )
        s1.on_next(5)
        sleep(0.05)
        self.assertEqual(results, ["val:5"])
        sub.cancel()

    def test_combine_latest_cancellation_disposes_inner(self) -> None:
        s1_ps = PublishSubject(int)
        s2_ps = PublishSubject(str)
        s1_dispose_called = Value(False)
        s2_dispose_called = Value(False)

        # Wrap subjects to track disposal
        class DisposableSrc(Observable[Any]):
            def __init__(self, actual_src: Subscribable[Any], flag: Value[bool]):
                super().__init__()
                self._actual_src = actual_src
                self._flag = flag

            def subscribe(
                self,
                on_next=None,
                on_error=None,
                on_completed=None,
                on_dispose=None,
                asynchronous=False,
                backpressure=None,
            ):
                def _dispose_hook():
                    self._flag.set(True)
                    if on_dispose:
                        on_dispose()

                return self._actual_src.subscribe(
                    on_next,
                    on_error,
                    on_completed,
                    _dispose_hook,
                    asynchronous,
                    backpressure,
                )

        disposable_s1 = DisposableSrc(s1_ps, s1_dispose_called)
        disposable_s2 = DisposableSrc(s2_ps, s2_dispose_called)

        combined_sub = RX.combine_latest(
            tuple[Any, Any], disposable_s1, disposable_s2
        ).subscribe(asynchronous=True)

        s1_ps.on_next(1)  # Ensure subscriptions are active
        s2_ps.on_next("a")
        sleep(0.05)

        combined_sub.cancel()
        sleep(0.05)

        self.assertTrue(s1_dispose_called.get(), "s1 should have been disposed")
        self.assertTrue(s2_dispose_called.get(), "s2 should have been disposed")

        s1_ps.on_completed(None)  # Clean up subjects
        s2_ps.on_completed(None)



from baseTest import BaseTestCase
from jstreams.mapper import Mapper, MapperWith, mapper_of, mapper_with_of


class TestMapper(BaseTestCase):
    def test_mapper(self):
        def fn(x):
            return x + 1

        self.assertEqual(mapper_of(fn), Mapper.of(fn))
        self.assertEqual(mapper_of(Mapper.of(fn)), Mapper.of(fn))
        self.assertNotEqual(mapper_of(fn), lambda x: x + 1)

    def test_mapper_with(self):
        def fn(x, y):
            return x + y

        self.assertEqual(mapper_with_of(fn), MapperWith.of(fn))
        self.assertEqual(mapper_with_of(MapperWith.of(fn)), MapperWith.of(fn))
        self.assertNotEqual(mapper_with_of(fn), lambda x, y: x + y)



from typing import Any


class NoOpCls:
    """
    A No-Operation object that does nothing when its methods are called
    and returns itself for method chaining.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the Noop object (does nothing)."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Allows the Noop object to be called as a function (does nothing)."""
        return self

    def __getattr__(self, name: str) -> Any:
        """Returns itself for any attribute access, enabling method chaining."""
        return self

    def __setattr__(self, name: str, value: Any) -> None:
        """Ignores attribute setting."""

    def __delattr__(self, name: str) -> None:
        """Ignores attribute deletion."""

    def __repr__(self) -> str:
        """Returns a string representation of the Noop object."""
        return "<Noop>"

    def __str__(self) -> str:
        """Returns a string representation of the Noop object."""
        return "<Noop>"

    def __bool__(self) -> bool:
        """Returns False, as a Noop object has no meaningful value."""
        return False

    def __eq__(self, other: Any) -> bool:
        """Returns True if the other object is also a Noop, False otherwise."""
        return isinstance(other, NoOpCls)

    def __ne__(self, other: Any) -> bool:
        """Returns True if the other object is not a Noop, False otherwise."""
        return not isinstance(other, NoOpCls)

    def __enter__(self) -> Any:
        return self


NoOp = NoOpCls()


def noop() -> NoOpCls:
    return NoOp



import sys
from datetime import date, datetime
from enum import Enum
import re
from typing import (
    Any,
    Callable,
    Optional,
    TypeVar,
    Iterable,
    Union,
    Protocol,
    runtime_checkable,
    get_type_hints,
    get_origin,
    get_args,
)
import inspect
from uuid import UUID

# TypeVar to represent the class being decorated.
_T = TypeVar("_T")
# Covariant TypeVar for SerializableObject protocol
_T_co = TypeVar("_T_co", covariant=True)

# Pre-compiled regex for _camel_to_snake for performance
_CAMEL_TO_SNAKE_PAT1 = re.compile(r"(.)([A-Z][a-z]+)")
_CAMEL_TO_SNAKE_PAT2 = re.compile(r"__([A-Z])")
_CAMEL_TO_SNAKE_PAT3 = re.compile(r"([a-z0-9])([A-Z])")

# Caches for _deserialize_value to speed up repeated type checks
_ENUM_TYPE_CACHE: dict[type, bool] = {}
_HAS_CALLABLE_FROM_DICT_CACHE: dict[type, bool] = {}


def _snake_to_camel(snake_str: str) -> str:
    """Converts a snake_case string to camelCase."""
    if not snake_str:
        return snake_str

    # Optimized handling of leading underscores
    first_char_index = 0
    s_len = len(snake_str)
    while first_char_index < s_len and snake_str[first_char_index] == "_":
        first_char_index += 1

    leading_underscores = snake_str[:first_char_index]
    name_part = snake_str[first_char_index:]

    if not name_part:  # True if original string was all underscores or empty
        return snake_str
    # If snake_str was empty, it's returned at the top.
    # If it was all underscores (e.g., "___"), leading_underscores="___", name_part="".
    # It will correctly return "___".

    components = name_part.split("_")
    camel_case_part = components[0] + "".join(x.title() for x in components[1:])
    return leading_underscores + camel_case_part


def _camel_to_snake(camel_str: str) -> str:
    """Converts a camelCase string to snake_case."""
    if not camel_str:
        return camel_str
    name = _CAMEL_TO_SNAKE_PAT1.sub(r"\1_\2", camel_str)
    name = _CAMEL_TO_SNAKE_PAT2.sub(r"_\1", name)
    name = _CAMEL_TO_SNAKE_PAT3.sub(r"\1_\2", name)
    return name.lower()


# Define a Protocol for objects that have to_dict and from_dict methods.
# Define a Protocol for objects that have a to_dict method.
# This helps in type checking the duck-typed call to value.to_dict().
@runtime_checkable
class SerializableObject(Protocol[_T_co]):
    def to_dict(self) -> dict[str, Any]: ...
    @classmethod
    def from_dict(cls: type[_T_co], data: dict[str, Any]) -> _T_co: ...


def _process_value(value: Any) -> Any:
    """Helper function to recursively process values for serialization."""
    # If the object conforms to SerializableObject, isinstance check is sufficient due to @runtime_checkable

    if isinstance(value, SerializableObject):
        # If the object conforms to SerializableObject and has a callable to_dict
        return value.to_dict()
    if isinstance(value, (datetime, date)):
        return value.isoformat()
    if isinstance(value, UUID):
        return str(value)
    if isinstance(value, Enum):
        return value.value  # Serialize Enums to their value
    if isinstance(value, list):
        return [_process_value(item) for item in value]
    if isinstance(value, tuple):
        # Preserve tuple type by creating a new tuple with processed items
        return tuple(_process_value(item) for item in value)
    if isinstance(value, dict):
        return {k: _process_value(v) for k, v in value.items()}
    # Basic types (int, str, float, bool, None) and other unhandled types are returned as-is.
    return value


def _deserialize_value(target_type: Any, data_value: Any) -> Any:
    """
    Helper function to recursively deserialize a data value to a target type.
    """
    if target_type is Any:  # No specific type hint, return as is
        return data_value

    origin = get_origin(target_type)
    actual_type_to_check = origin or target_type

    # 1. Handle specific known types first
    if actual_type_to_check is datetime and isinstance(data_value, str):
        try:
            return datetime.fromisoformat(data_value)
        except ValueError:
            pass  # Fall through if not a valid ISO format string
    if actual_type_to_check is date and isinstance(data_value, str):
        try:
            return date.fromisoformat(data_value)
        except ValueError:
            pass
    if actual_type_to_check is UUID and isinstance(data_value, str):
        try:
            return UUID(data_value)
        except ValueError:
            pass
    if inspect.isclass(actual_type_to_check) and issubclass(actual_type_to_check, Enum):
        # Optimized Enum check using cache
        is_enum_class = _ENUM_TYPE_CACHE.get(actual_type_to_check)
        if is_enum_class is None:
            is_enum_class = inspect.isclass(actual_type_to_check) and issubclass(
                actual_type_to_check, Enum
            )
            _ENUM_TYPE_CACHE[actual_type_to_check] = is_enum_class

        if is_enum_class:
            try:
                # Attempt to create enum member from value
                return actual_type_to_check(data_value)
            except ValueError:
                # If data_value is not a valid value for the enum,
                # it might be the enum member name (less common for JSON).
                pass
    origin = get_origin(target_type)
    args = get_args(target_type)

    if origin is Union:  # Handles Optional[X] (Union[X, NoneType]) and other Unions
        if data_value is None:
            # If None is a valid type in the Union (e.g., Optional), return None.
            if any(arg is type(None) for arg in args):
                return None
            # If None is not allowed by the Union but data is None, this is a potential mismatch.
            # For now, we let it fall through; specific error handling could be added.

        # Try to deserialize with non-None types in the Union.
        for arg_type in args:
            if arg_type is type(None):
                continue
            try:
                # Attempt to deserialize with this type from the Union.
                return _deserialize_value(arg_type, data_value)
            except (TypeError, ValueError, AttributeError, KeyError):
                # If deserialization with this arg_type fails, try the next one.
                continue

        # If no type in Union matches or successfully deserializes,
        # and data_value was not None (or None was not allowed), return raw data_value.
        # This might be the correct behavior if data_value is already of a compatible simple type
        # that wasn't explicitly in the Union's args but is assignable.
        if data_value is not None or not any(arg is type(None) for arg in args):
            return data_value
        return None  # Should have been caught by 'if data_value is None' earlier if None was allowed

    if origin in (list, tuple) and args:  # Handles list[X], tuple[X, Y, ...]
        item_type = args[0]  # Assuming list[X] or tuple[X, ...]
        if isinstance(data_value, list):  # Expecting a list from JSON-like data
            processed_list = [
                _deserialize_value(item_type, item) for item in data_value
            ]
            return processed_list if origin is list else origin(processed_list)
        # Data doesn't match expected list type, return as is or raise error.
        return data_value  # Or raise TypeError(f"Expected list for {target_type}, got {type(data_value)}")

    if origin is dict and args and len(args) == 2:  # Handles dict[KeyType, ValueType]
        # key_type = args[0] # Assuming key_type is simple (e.g. str) for JSON-like dicts
        value_type = args[1]
        if isinstance(data_value, dict):
            return {k: _deserialize_value(value_type, v) for k, v in data_value.items()}
        # Data doesn't match expected dict type
        return data_value  # Or raise TypeError

    # Check for a class that has from_dict (could be target_type itself if not a generic)
    # This handles direct SerializableObject types.
    # Optimized check for from_dict capability using cache
    has_callable_from_dict = _HAS_CALLABLE_FROM_DICT_CACHE.get(actual_type_to_check)
    if has_callable_from_dict is None:
        if inspect.isclass(
            actual_type_to_check
        ):  # hasattr/getattr are safer on classes
            from_dict_method = getattr(actual_type_to_check, "from_dict", None)
            has_callable_from_dict = callable(from_dict_method)
        else:
            has_callable_from_dict = (
                False  # Not a class, so can't have classmethod from_dict
            )
        _HAS_CALLABLE_FROM_DICT_CACHE[actual_type_to_check] = has_callable_from_dict

    if has_callable_from_dict and isinstance(data_value, dict):
        # actual_type_to_check is known to have a callable from_dict
        return actual_type_to_check.from_dict(data_value)

    # If data_value is already an instance of the target type (after considering origin)
    if actual_type_to_check is not Any and isinstance(data_value, actual_type_to_check):
        return data_value

    # Attempt basic type coercion if target_type is a basic type and data_value is not already that type
    if actual_type_to_check in (int, float, str, bool) and not isinstance(
        data_value, actual_type_to_check
    ):
        try:
            return actual_type_to_check(data_value)
        except (ValueError, TypeError):
            # Coercion failed, fall through to return original data_value
            pass
    return data_value


def json_standard_serializable(
    ignore_unknown_fields: bool = True,
    aliases: Optional[dict[str, str]] = None,
    omit_none: bool = False,
    custom_serializers: Optional[dict[str, Callable[[Any], Any]]] = None,
    custom_deserializers: Optional[dict[str, Callable[[Any], Any]]] = None,
    post_deserialize_hook_name: Optional[str] = "__post_deserialize__",
) -> Callable[[type[_T]], type[_T]]:
    """
    Alias for json_serializable with translate_snake_to_camel set to True.
    This decorator is used to create a JSON serializable class.

    Args:
        ignore_unknown_fields (bool, optional): _description_. Defaults to True.
        aliases (Optional[dict[str, str]], optional): _description_. Defaults to None.
        omit_none (bool, optional): _description_. Defaults to False.
        custom_serializers (Optional[dict[str, Callable[[Any], Any]]], optional): _description_. Defaults to None.
        custom_deserializers (Optional[dict[str, Callable[[Any], Any]]], optional): _description_. Defaults to None.
        post_deserialize_hook_name (Optional[str], optional): _description_. Defaults to "__post_deserialize__".

    Returns:
        Callable[[type[_T]], type[_T]]: _description_
    """
    return json_serializable(
        ignore_unknown_fields=ignore_unknown_fields,
        aliases=aliases,
        omit_none=omit_none,
        custom_serializers=custom_serializers,
        custom_deserializers=custom_deserializers,
        post_deserialize_hook_name=post_deserialize_hook_name,
        translate_snake_to_camel=True,
    )


def json_serializable(
    ignore_unknown_fields: bool = True,
    aliases: Optional[dict[str, str]] = None,
    omit_none: bool = False,
    custom_serializers: Optional[dict[str, Callable[[Any], Any]]] = None,
    custom_deserializers: Optional[dict[str, Callable[[Any], Any]]] = None,
    post_deserialize_hook_name: Optional[str] = "__post_deserialize__",
    translate_snake_to_camel: bool = False,
) -> Callable[[type[_T]], type[_T]]:
    """
    A class decorator that adds a to_dict() method to the decorated class.
    This method serializes an instance into a dictionary, including attributes
    from both __dict__ (standard instance attributes) and __slots__ (if used).
    The decorator also provides a from_dict() class method for deserialization.
    This method takes a dictionary and populates the instance's attributes
    based on the provided data. It uses type hints to determine the expected
    types of the attributes and attempts to deserialize the data accordingly.

    Serialization is recursive:
    - Objects that have a 'to_dict' method (e.g., other @serializable instances)
      are serialized by calling their to_dict() method.
    - Lists and tuples are iterated, and their items are processed recursively.
    - Dictionaries are iterated, and their values are processed recursively.
    - Supports field aliasing for serialization and deserialization.
    - Can omit fields with None values during serialization.
    - Supports custom serializer functions for specific fields.
    - Supports custom deserializer functions for specific fields.
    - Can call a post-deserialization hook method on the instance.
    - Enhanced to handle common types like datetime, date, Enum, UUID.

    By convention, attributes are excluded from serialization if their names:
    - Start with a single underscore (e.g., _protected_attribute).
    - Start with double underscores (e.g., __private_attribute, which undergoes
      name mangling to _ClassName__private_attribute and is thus also excluded).
    This helps in respecting encapsulation and not serializing internal state.
    It also supports:
    - Aliasing of attributes during deserialization.
    - Handling of unknown fields (fields not defined in the class).
    - Custom deserializer functions for specific fields.
    - Post-deserialization hook method that can be called after deserialization.
    The decorator can be applied to any class, including those that use
    dataclasses, namedtuples, or other custom structures.

    Args:
        ignore_unknown_fields (bool): If True, unknown fields in the input data
            will be ignored during deserialization.
        aliases (Optional[dict[str, str]]): A dictionary mapping attribute names
            to their aliases. This allows for renaming attributes during
            serialization/deserialization.  Defaults to None.
        omit_none (bool): If True, attributes with None values will be omitted
            from the serialized output. Defaults to False.
        custom_serializers (Optional[dict[str, Callable[[Any], Any]]]): A dictionary
            mapping attribute names to custom serializer functions. These functions
            will be called to serialize the corresponding attribute. Defaults to None.
        custom_deserializers (Optional[dict[str, Callable[[Any], Any]]]): A dictionary
            mapping attribute names to custom deserializer functions. These functions
            will be called to deserialize the corresponding attribute. Defaults to None.
        post_deserialize_hook_name (Optional[str]): The name of a method to be called
            after deserialization. This method should be defined in the class and
            will be called with no arguments. Defaults to "__post_deserialize__".
        translate_snake_to_camel (bool): If True, attribute names will be converted
            from snake_case to camelCase during serialization. Defaults to False.
    """

    def decorator(cls: type[_T]) -> type[_T]:
        # --- Caching introspection results at decoration time ---
        # Use a unique prefix for cached attributes to avoid potential collisions.
        cached_type_hints_attr = f"_jstreams_cached_type_hints_{cls.__name__}"
        cached_init_params_attr = f"_jstreams_cached_init_params_{cls.__name__}"
        cached_reverse_aliases_attr = f"_jstreams_cached_reverse_aliases_{cls.__name__}"
        cached_slots_attr = f"_jstreams_cached_slots_{cls.__name__}"

        # Cache type hints
        if not hasattr(cls, cached_type_hints_attr):
            try:
                hints = get_type_hints(cls)
                setattr(cls, cached_type_hints_attr, hints)
            except Exception:  # get_type_hints can fail in some edge cases
                setattr(cls, cached_type_hints_attr, {})

        # Cache __init__ parameters
        if not hasattr(cls, cached_init_params_attr):
            try:
                init_sig = inspect.signature(cls.__init__)
                setattr(cls, cached_init_params_attr, init_sig.parameters)
            except (ValueError, TypeError):  # Not inspectable
                setattr(cls, cached_init_params_attr, {})

        # Cache reverse aliases
        if not hasattr(cls, cached_reverse_aliases_attr):
            setattr(
                cls,
                cached_reverse_aliases_attr,
                {v: k for k, v in (aliases or {}).items()},
            )

        # Cache __slots__ as a set for efficient lookup
        if not hasattr(cls, cached_slots_attr):
            if hasattr(cls, "__slots__"):
                defined_slots = cls.__slots__  # type: ignore[attr-defined]
                if isinstance(defined_slots, str):
                    setattr(cls, cached_slots_attr, {defined_slots})
                else:  # Assuming iterable
                    setattr(cls, cached_slots_attr, set(defined_slots))
            else:
                setattr(cls, cached_slots_attr, set())  # Empty set if no slots

        # --- End Caching ---

        def to_dict(self: _T) -> dict[str, Any]:
            return _to_dict_convert_name(self, convert_names=True)

        def _to_dict_convert_name(self: _T, convert_names: bool) -> dict[str, Any]:
            serialized_data: dict[str, Any] = {}
            aliases_map_local = aliases or {}
            custom_serializers_map_local = custom_serializers or {}

            # This map will store all potential attributes to serialize,
            # gathered from __dict__ and __slots__.
            attributes_map: dict[str, Any] = {}

            # 1. Gather attributes from __dict__ if it exists
            if hasattr(self, "__dict__"):
                attributes_map.update(self.__dict__)

            # 2. Gather attributes from __slots__ if the class defines them
            # getattr() is used to fetch slot values. This correctly handles
            # attributes defined only in __slots__ and also cases where __slots__
            # might include '__dict__' (allowing both slotted and dynamic attributes).
            if hasattr(cls, "__slots__"):
                # cls.__slots__ can be a string or an iterable of strings
                defined_slot_names: Union[str, Iterable[str]] = cls.__slots__  # type: ignore[attr-defined]
                actual_slot_names: Iterable[str]

                if isinstance(defined_slot_names, str):
                    actual_slot_names = [defined_slot_names]
                else:
                    actual_slot_names = defined_slot_names

                for slot_name in actual_slot_names:
                    # Special slots like __dict__ and __weakref__ are part of Python's object model
                    # and not typically considered data attributes for serialization.
                    # __dict__ contents are already handled above. __weakref__ is for weak references.
                    if slot_name in ("__dict__", "__weakref__"):
                        continue
                    try:
                        # Fetch the value of the slot from the instance
                        attributes_map[slot_name] = getattr(self, slot_name)
                    except AttributeError:
                        # This can happen if a slot is defined but not yet assigned a value
                        # on this particular instance, or if it's a complex descriptor.
                        # In such cases, we'll skip it for serialization.
                        continue  # Use continue to be explicit about skipping

            # 3. Gather attributes defined purely by __annotations__ that exist on the instance
            # This handles cases where an attribute might be defined with a type hint
            # but not explicitly in __init__ or __slots__, but is set later on the instance.
            if hasattr(cls, "__annotations__"):
                for key in cls.__annotations__:
                    # If the attribute is already in __dict__ or __slots__, we've already got it.
                    # We only need to check annotations for attributes *not* yet in the map.
                    if key not in attributes_map:
                        # Try to get the value from the instance.
                        # If it exists, add it to the map. If not, skip it.
                        try:
                            attributes_map[key] = getattr(self, key)
                        except AttributeError:
                            # Attribute defined in annotations but not set on this instance.
                            # Skip it for serialization.
                            continue

            for attr_name, raw_value in attributes_map.items():
                if not attr_name.startswith(
                    "_"
                ):  # Exclude "private" or "protected" attributes
                    processed_value: Any
                    if attr_name in custom_serializers_map_local:
                        processed_value = custom_serializers_map_local[attr_name](
                            raw_value
                        )
                    else:
                        processed_value = _process_value(raw_value)

                    if omit_none and processed_value is None:
                        continue

                    output_key = aliases_map_local.get(attr_name, attr_name)
                    if translate_snake_to_camel and convert_names:
                        output_key = _snake_to_camel(output_key)
                    serialized_data[output_key] = processed_value

            return serialized_data

        def from_dict(cls_target: type[_T], data: dict[str, Any]) -> _T:
            """
            Creates an instance of the class from a dictionary.
            Recursively deserializes nested objects based on type hints.
            """
            # Retrieve cached introspection results
            all_type_hints = getattr(cls_target, cached_type_hints_attr)
            init_params = getattr(cls_target, cached_init_params_attr)
            reverse_aliases_map = getattr(cls_target, cached_reverse_aliases_attr)
            cls_slots = getattr(
                cls_target, cached_slots_attr
            )  # Retrieve cached slots set

            custom_deserializers_map_local = custom_deserializers or {}

            init_kwargs: dict[str, Any] = {}
            extra_data: dict[str, Any] = {}  # For data not used in __init__

            # Prepare arguments for __init__ by deserializing them
            for key_from_data, value_from_data in data.items():
                sanitized_key = key_from_data
                if translate_snake_to_camel:
                    sanitized_key = _camel_to_snake(key_from_data)
                attr_name = reverse_aliases_map.get(sanitized_key, sanitized_key)

                if attr_name in init_params and init_params[attr_name].name != "self":
                    param = init_params[attr_name]
                    # Determine the type hint for the __init__ parameter
                    param_type_hint = param.annotation
                    if param_type_hint is inspect.Parameter.empty:
                        # If __init__ param has no type hint, try class-level hint
                        param_type_hint = all_type_hints.get(attr_name, Any)

                    if attr_name in custom_deserializers_map_local:
                        init_kwargs[attr_name] = custom_deserializers_map_local[
                            attr_name
                        ](value_from_data)

                    else:
                        init_kwargs[attr_name] = _deserialize_value(
                            param_type_hint, value_from_data
                        )
                else:
                    # Store with original (potentially aliased) key if it's not an init param,
                    # or de-aliased name if it was an init param but not used (e.g. **kwargs)
                    # For simplicity, always use de-aliased name for extra_data keys.
                    extra_data[attr_name] = value_from_data

            # Instantiate the object using prepared __init__ arguments
            # This assumes __init__ can handle the provided kwargs.
            # If required __init__ args are missing from data and have no defaults,
            # this will (correctly) raise a TypeError.
            instance = cls_target(**init_kwargs)

            # Set any remaining attributes from `extra_data` using setattr
            for attr_name_from_extra, original_value_from_data in extra_data.items():
                is_known_attribute = False  # Known via slots or type hints
                if attr_name_from_extra in cls_slots:  # Use cached slots
                    is_known_attribute = True

                if not is_known_attribute and attr_name_from_extra in all_type_hints:
                    is_known_attribute = True

                if is_known_attribute:
                    attr_type_hint = all_type_hints.get(attr_name_from_extra, Any)
                    final_value: Any
                    if attr_name_from_extra in custom_deserializers_map_local:
                        final_value = custom_deserializers_map_local[
                            attr_name_from_extra
                        ](original_value_from_data)
                    else:
                        final_value = _deserialize_value(
                            attr_type_hint, original_value_from_data
                        )
                    try:
                        setattr(instance, attr_name_from_extra, final_value)
                    except (
                        AttributeError
                    ):  # e.g., property without setter, or __slots__ issue
                        # Optionally log: f"Warning: Could not set attribute '{key}' on '{cls_target.__name__}'"
                        pass
                elif (
                    not ignore_unknown_fields
                ):  # Attribute is not known, but we are not ignoring unknown fields
                    # Attempt to set it. Deserialize with Any, which processes collections/SerializableObjects
                    # but otherwise uses the value as-is.
                    value_to_set = _deserialize_value(Any, original_value_from_data)
                    try:
                        setattr(instance, attr_name_from_extra, value_to_set)
                    except AttributeError:
                        pass

            # Call post-deserialization hook if it exists
            if post_deserialize_hook_name and hasattr(
                instance, post_deserialize_hook_name
            ):
                hook_method = getattr(instance, post_deserialize_hook_name)
                if callable(hook_method):
                    try:
                        hook_method()
                    except Exception:
                        # Optionally log or handle exception from hook
                        # print(f"Warning: Error in post_deserialize_hook '{post_deserialize_hook_name}': {e}")
                        pass

            return instance

        def __eq__(self: _T, other: Any) -> bool:
            """
            Compares this instance with another object for equality.
            Two instances are considered equal if they are of the same type
            and their `to_dict()` representations are identical.
            """
            if not isinstance(other, self.__class__):
                return NotImplemented  # Or False, NotImplemented is more idiomatic for __eq__

            # Both self and other should have to_dict if decorated by serializable
            # and self.__class__ is the same as other.__class__
            self_dict = self._to_dict_convert_name(False)  # type: ignore[attr-defined]
            other_dict = other._to_dict_convert_name(False)  # type: ignore[attr-defined]
            return self_dict == other_dict  # type: ignore[no-any-return]

        def __str__(self: _T) -> str:
            """
            Returns a string representation of the instance.
            This is useful for debugging and logging.
            """
            return f"{self.__class__.__name__}({self._to_dict_convert_name(False)})"  # type: ignore[attr-defined]

        # Add the to_dict method to the class.
        setattr(cls, "to_dict", to_dict)
        setattr(cls, "_to_dict_convert_name", _to_dict_convert_name)
        setattr(cls, "from_dict", classmethod(from_dict))
        setattr(cls, "__eq__", __eq__)
        setattr(cls, "__str__", __str__)
        setattr(cls, "__repr__", __str__)
        return cls

    return decorator


def json_deserialize(class_type: type[_T], data: dict[str, Any]) -> _T:
    """
    Deserialize a dictionary into an instance of the specified class type.
    This function is a convenience wrapper around the from_dict method of the class.
    """
    if not hasattr(class_type, "from_dict"):
        raise TypeError(f"{class_type.__name__} does not have a from_dict method.")
    return class_type.from_dict(data)  # type: ignore


def json_deserialize_list(class_type: type[_T], data: list[dict[str, Any]]) -> list[_T]:
    """
    Deserialize a list of dictionaries into a list of instances of the specified class type.
    """
    if not hasattr(class_type, "from_dict"):
        raise TypeError(f"{class_type.__name__} does not have a from_dict method.")
    for item in data:
        if not isinstance(item, dict):
            raise TypeError(f"Expected a dictionary, got {type(item)}")

    return [class_type.from_dict(item) for item in data]  # type: ignore


def json_serialize(obj: Any) -> dict[str, Any]:
    """
    Serialize an object into a dictionary.
    This function is a convenience wrapper around the to_dict method of the object.
    """
    if not hasattr(obj, "to_dict"):
        raise TypeError(f"{obj.__class__.__name__} does not have a to_dict method.")
    return obj.to_dict()  # type: ignore


def json_serialize_list(obj: list[Any]) -> list[dict[str, Any]]:
    """
    Serialize an list of objects into a list of dictionaries.
    This function is a convenience wrapper around the to_dict method of the object.
    """
    for item in obj:
        if not hasattr(item, "to_dict"):
            raise TypeError(
                f"{item.__class__.__name__} does not have a to_dict method."
            )
    return [item.to_dict() for item in obj]


if sys.version_info >= (3, 10):
    from typing import ParamSpec

    P = ParamSpec("P")
else:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")
# TypeVar for the original return type of a callable
R = TypeVar("R")


def json_serialize_return() -> Callable[  # Type of the decorator factory
    [Callable[P, R]],  # It accepts a callable func(P) -> R
    Callable[P, dict[str, Any]],  # It returns a new callable func(P) -> dict
]:
    """
    Decorator factory to serialize the return value of a function using json_serialize.
    The decorated function will have its return value processed by json_serialize.
    """

    def decorator(func: Callable[P, R]) -> Callable[P, dict[str, Any]]:
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> dict[str, Any]:
            result = func(*args, **kwargs)
            return json_serialize(result)

        return wrapper

    return decorator


def json_serialize_return_list() -> Callable[  # Type of the decorator factory
    [Callable[P, list[R]]],  # It accepts a callable func(P) -> list[R]
    Callable[
        P, list[dict[str, Any]]
    ],  # It returns a new callable func(P) -> list[dict]
]:
    """
    Decorator factory to serialize the return value of a function using json_serialize.
    The decorated function will have its return value processed by json_serialize.
    """

    def decorator(func: Callable[P, list[R]]) -> Callable[P, list[dict[str, Any]]]:
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> list[dict[str, Any]]:
            result = func(*args, **kwargs)
            return [json_serialize(item) for item in result]

        return wrapper

    return decorator



import threading
import time
from typing import Optional
from baseTest import BaseTestCase
from jstreams.annotations import (
    all_args,
    getter,
    locked,
    required_args,
    setter,
    builder,
)


# --- Test Subject Class for @locked ---
class UnsafeCounter:
    """A simple counter class, intentionally not thread-safe."""

    def __init__(self, initial_value: int = 0) -> None:
        # print(f"Initializing UnsafeCounter with {initial_value}") # Keep commented out unless debugging
        self.value = initial_value
        self._internal_state = "initialized"  # Example protected member

    def increment(self, amount: int = 1) -> None:
        """Increments the counter (simulates work)."""
        # print(f"{threading.current_thread().name}: Reading value {self.value}") # Keep commented out
        current_value = self.value  # Read
        time.sleep(0.01)  # Simulate work/potential race condition
        self.value = current_value + amount  # Write
        # print(f"{threading.current_thread().name}: Incremented to {self.value}") # Keep commented out

    def get_value(self) -> int:
        """Returns the current value."""
        # print(f"{threading.current_thread().name}: Getting value {self.value}") # Keep commented out
        return self.value

    def complex_operation(self, amount1: int, amount2: int) -> int:
        """Calls increment multiple times to test reentrancy."""
        self.increment(amount1)
        time.sleep(0.005)
        self.increment(amount2)
        return self.get_value()

    def get_internal_state(self) -> str:
        """Accesses a 'protected' member."""
        return self._internal_state

    def __str__(self) -> str:
        return f"UnsafeCounter(value={self.value})"

    def __repr__(self) -> str:
        return f"UnsafeCounter({self.value})"


class InitErrorCounter:
    def __init__(self) -> None:
        raise ValueError("Failed to initialize")


class TestAnnotations(BaseTestCase):
    def test_getter(self) -> None:
        @getter()
        class Test:
            var1: str
            var2: int
            var3: str
            _var_private: int = 1

        @getter()
        class OtherTest:
            other_var: str
            other_var2: float

        test_instance = Test()
        test_instance.var1 = "value1"
        test_instance.var2 = 123
        test_instance.var3 = "value3"

        self.assertEqual(test_instance.get_var1(), "value1")
        self.assertEqual(test_instance.get_var2(), 123)
        self.assertEqual(test_instance.get_var3(), "value3")

        other_test_instance = OtherTest()
        other_test_instance.other_var = "test"
        other_test_instance.other_var2 = 2.3

        self.assertEqual(other_test_instance.get_other_var(), "test")
        self.assertEqual(other_test_instance.get_other_var2(), 2.3)

        self.assertRaises(AttributeError, lambda: test_instance.get_unknown())
        self.assertRaises(AttributeError, lambda: test_instance.get__var_private())

    def test_setter(self) -> None:
        @setter()
        class Test:
            var1: str
            var2: int
            var3: str
            _var_private: int = 1

        @setter()
        class OtherTest:
            other_var: str
            other_var2: float

        test_instance = Test()
        test_instance.set_var1("value1")
        test_instance.set_var2(123)
        test_instance.set_var3("value3")

        self.assertEqual(test_instance.var1, "value1")
        self.assertEqual(test_instance.var2, 123)
        self.assertEqual(test_instance.var3, "value3")

        other_test_instance = OtherTest()
        other_test_instance.set_other_var("test")
        other_test_instance.set_other_var2(2.3)

        self.assertEqual(other_test_instance.other_var, "test")
        self.assertEqual(other_test_instance.other_var2, 2.3)
        self.assertRaises(AttributeError, lambda: test_instance.set_unknown(1))
        self.assertRaises(AttributeError, lambda: test_instance.set__var_private(1))

    def test_builder(self) -> None:
        @builder()
        class Test:
            var1: str
            var2: int
            var3: str
            _var_private: int = 1

        @builder()
        class OtherTest:
            other_var: str
            other_var2: float

        test_instance: Test = (
            Test.builder()
            .with_var1("value1")
            .with_var2(123)
            .with_var3("value3")
            .build()
        )

        self.assertEqual(test_instance.var1, "value1")
        self.assertEqual(test_instance.var2, 123)
        self.assertEqual(test_instance.var3, "value3")

        other_test_instance: OtherTest = (
            OtherTest.builder().with_other_var("test").with_other_var2(2.3).build()
        )

        self.assertEqual(other_test_instance.other_var, "test")
        self.assertEqual(other_test_instance.other_var2, 2.3)
        self.assertRaises(
            AttributeError, lambda: Test.builder().with_unknown(1).build()
        )
        self.assertRaises(
            AttributeError, lambda: Test.builder().with__var_private(1).build()
        )

    def test_locked_single_thread(self) -> None:
        """Tests basic functionality of @locked in a single thread."""
        LockedCounter = locked()(UnsafeCounter)  # Apply decorator manually for testing
        counter = LockedCounter(10)

        self.assertEqual(counter.get_value(), 10)
        counter.increment(5)
        self.assertEqual(counter.get_value(), 15)
        counter.value = 50  # Test direct attribute setting
        self.assertEqual(counter.get_value(), 50)
        self.assertEqual(
            counter.get_internal_state(), "initialized"
        )  # Test reading protected-like attr

        # Test special methods
        self.assertTrue("ThreadSafeWrapper" in repr(counter))
        self.assertTrue("UnsafeCounter(value=50)" in str(counter))

    def test_locked_multi_thread_race_condition(self) -> None:
        """Tests if @locked prevents race conditions with multiple threads."""
        LockedCounter = locked()(UnsafeCounter)
        counter = LockedCounter(0)  # Shared instance

        num_threads = 10
        increments_per_thread = 20
        threads = []

        def worker() -> None:
            for _ in range(increments_per_thread):
                counter.increment()  # Access shared instance

        for i in range(num_threads):
            thread = threading.Thread(target=worker, name=f"Worker-{i + 1}")
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        expected_value = num_threads * increments_per_thread
        actual_value = counter.get_value()
        self.assertEqual(
            actual_value,
            expected_value,
            "Counter value should be correct after concurrent increments",
        )

    def test_locked_multi_thread_reentrancy(self) -> None:
        """Tests if RLock allows reentrant calls within the same thread."""
        LockedCounter = locked()(UnsafeCounter)
        counter = LockedCounter(0)  # Shared instance

        num_threads = 5
        ops_per_thread = 10
        threads = []

        def worker() -> None:
            for _ in range(ops_per_thread):
                # This method calls increment internally, testing reentrancy
                counter.complex_operation(2, 3)  # Total increment of 5 per call

        for i in range(num_threads):
            thread = threading.Thread(target=worker, name=f"ReentrantWorker-{i + 1}")
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        expected_value = num_threads * ops_per_thread * 5  # 5 = 2 + 3
        actual_value = counter.get_value()
        self.assertEqual(
            actual_value,
            expected_value,
            "Counter value should be correct after concurrent complex operations",
        )

    def test_locked_attribute_errors(self) -> None:
        """Tests attribute access errors on the locked wrapper."""
        LockedCounter = locked()(UnsafeCounter)
        counter = LockedCounter(0)

        # Test getting non-existent attribute
        self.assertRaises(AttributeError, lambda: getattr(counter, "non_existent"))

    def test_locked_init_exception(self) -> None:
        """Tests that exceptions during original __init__ are propagated."""
        LockedInitErrorCounter = locked()(InitErrorCounter)

        with self.assertRaisesRegex(ValueError, "Failed to initialize"):
            LockedInitErrorCounter()  # Instantiation should raise the error from original __init__

    def test_locked_metadata(self) -> None:
        """Tests if the wrapper preserves some metadata."""

        @locked()
        class MyOriginalClass:
            """This is the original docstring."""

            pass

        self.assertEqual(MyOriginalClass.__name__, "ThreadSafeMyOriginalClass")
        self.assertIn(
            "Thread-safe wrapper around MyOriginalClass", MyOriginalClass.__doc__
        )
        self.assertIn("This is the original docstring.", MyOriginalClass.__doc__)

    def test_required_args(self) -> None:
        @required_args()
        class Test:
            a: int
            b: str
            c: Optional[int]
            d: Optional[str]

        t = Test.required(10, "test")
        self.assertEqual(t.a, 10)
        self.assertEqual(t.b, "test")
        self.assertIsNone(t.c)
        self.assertIsNone(t.d)

        self.assertRaises(TypeError, lambda: Test.required(10))
        self.assertRaises(TypeError, lambda: Test.required(10, "test", 2, "test2"))

    def test_all_args(self) -> None:
        @all_args()
        class Test:
            a: int
            b: str
            c: Optional[int] = None
            d: Optional[str] = None

        t = Test.all(10, "test", 2, "test2")
        self.assertEqual(t.a, 10)
        self.assertEqual(t.b, "test")
        self.assertEqual(t.c, 2)
        self.assertEqual(t.d, "test2")

        self.assertRaises(TypeError, lambda: Test.all())
        self.assertRaises(TypeError, lambda: Test.all(10))
        self.assertRaises(TypeError, lambda: Test.all(10, "str"))
        self.assertRaises(TypeError, lambda: Test.all(10, "str", 4))
        self.assertRaises(TypeError, lambda: Test.all(10, "test", 2, "test2", 2))



from abc import ABC, abstractmethod
import re
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Optional,
    TypeVar,
    Union,
    cast,
    Generic,
)

from collections.abc import Sized

T = TypeVar("T")
K = TypeVar("K")


class Predicate(ABC, Generic[T]):
    @abstractmethod
    def apply(self, value: T) -> bool:
        """
        Apply a condition to a given value.

        Args:
            value (T): The value

        Returns:
            bool: True if the value matches, False otherwise
        """

    def or_(self, other: Callable[[T], bool]) -> "Predicate[T]":
        return Predicate.of(lambda v: self.apply(v) or predicate_of(other).apply(v))

    def and_(self, other: Callable[[T], bool]) -> "Predicate[T]":
        return Predicate.of(lambda v: self.apply(v) and predicate_of(other).apply(v))

    def __call__(self, value: T) -> bool:
        return self.apply(value)

    @staticmethod
    def of(
        predicate: Callable[[T], bool],
    ) -> "Predicate[T]":
        """
        If the value passed is a predicate, it is returned without any changes.
        If a function is passed, it will be wrapped into a Predicate object.

        Args:
            predicate (Callable[[T], bool]): The predicate

        Returns:
            Predicate[T]: The produced predicate
        """
        if isinstance(predicate, Predicate):
            return predicate
        return _WrapPredicate(predicate)


class PredicateWith(ABC, Generic[T, K]):
    @abstractmethod
    def apply(self, value: T, with_value: K) -> bool:
        """
        Apply a condition to two given values.

        Args:
            value (T): The value
            with_value (K): The second value

        Returns:
            bool: True if the values matche the predicate, False otherwise
        """

    def or_(self, other: "PredicateWith[T, K]") -> "PredicateWith[T, K]":
        return predicate_with_of(lambda v, k: self.apply(v, k) or other.apply(v, k))

    def and_(self, other: "PredicateWith[T, K]") -> "PredicateWith[T, K]":
        return predicate_with_of(lambda v, k: self.apply(v, k) and other.apply(v, k))

    def __call__(self, value: T, with_value: K) -> bool:
        return self.apply(value, with_value)

    @staticmethod
    def of(
        predicate: Callable[[T, K], bool],
    ) -> "PredicateWith[T, K]":
        """
        If the value passed is a predicate, it is returned without any changes.
        If a function is passed, it will be wrapped into a Predicate object.

        Args:
            predicate (Callable[[T, K], bool]): The predicate

        Returns:
            PredicateWith[T, K]: The produced predicate
        """
        if isinstance(predicate, PredicateWith):
            return predicate
        return _WrapPredicateWith(predicate)


class _WrapPredicate(Predicate[T]):
    __slots__ = ("__predicate_fn",)

    def __init__(self, fn: Callable[[T], bool]) -> None:
        self.__predicate_fn = fn

    def apply(self, value: T) -> bool:
        return self.__predicate_fn(value)


class _WrapPredicateWith(PredicateWith[T, K]):
    __slots__ = ("__predicate_fn",)

    def __init__(self, fn: Callable[[T, K], bool]) -> None:
        self.__predicate_fn = fn

    def apply(self, value: T, with_value: K) -> bool:
        return self.__predicate_fn(value, with_value)


def predicate_of(predicate: Callable[[T], bool]) -> Predicate[T]:
    """
    If the value passed is a predicate, it is returned without any changes.
    If a function is passed, it will be wrapped into a Predicate object.

    Args:
        predicate (Callable[[T], bool]): The predicate

    Returns:
        Predicate[T]: The produced predicate
    """
    return Predicate.of(predicate)


def predicate_with_of(
    predicate: Callable[[T, K], bool],
) -> PredicateWith[T, K]:
    """
    If the value passed is a predicate, it is returned without any changes.
    If a function is passed, it will be wrapped into a Predicate object.

    Args:
        predicate (Callable[[T, K], bool]): The predicate

    Returns:
        PredicateWith[T, K]: The produced predicate
    """
    return PredicateWith.of(predicate)


def is_true(var: bool) -> bool:
    """
    Returns the same value. Meant to be used as a predicate for filtering.

    Args:
        var (bool): The value

    Returns:
        bool: The same value
    """
    return var


def is_false(var: bool) -> bool:
    """
    Returns the negated value.

    Args:
        var (bool): The value

    Returns:
        bool: the negated value
    """
    return not var


def is_none(val: Any) -> bool:
    """
    Equivalent to val is None. Meant to be used as a predicate.

    Args:
        val (Any): The value

    Returns:
        bool: True if None, False otherwise
    """
    return val is None


def is_in(it: Iterable[Any]) -> Predicate[Any]:
    """
    Predicate to check if a value is contained in an iterable.
    Usage: is_in(check_in_this_list)(find_this_item)
    Usage with Opt: Opt(val).filter(is_in(my_list))

    Args:
        it (Iterable[Any]): The iterable to check within.

    Returns:
        Predicate[Any]: The predicate.
    """

    # Efficiency depends on the type of 'it' (set/dict is O(1), list/tuple is O(n))
    def wrap(elem: Any) -> bool:
        return elem in it

    return Predicate.of(wrap)


def is_not_in(it: Iterable[Any]) -> Predicate[Any]:
    """
    Predicate to check if a value is not contained in an iterable.
    Usage: is_not_in(check_in_this_list)(find_this_item)
    Usage with Opt: Opt(val).filter(is_not_in(my_list))

    Args:
        it (Iterable[Any]): The iterable to check within.

    Returns:
        Predicate[Any]: The predicate.
    """
    # Reuses is_in and not_ for conciseness and correctness
    return not_(is_in(it))


def equals(obj: T) -> Predicate[T]:
    """
    Predicate to check if a value equals another value. Handles None correctly.
    Usage: equals(object_to_compare_to)(my_object)
    Usage with Opt: Opt(my_object).filter(equals(object_to_compare_to))

    Args:
        obj (T): The object to compare to.

    Returns:
        Predicate[T]: The predicate.
    """

    def wrap(other: T) -> bool:
        # Handles None comparison explicitly
        return (obj is None and other is None) or (obj == other)

    return Predicate.of(wrap)


def not_equals(obj: Any) -> Predicate[Any]:
    """
    Predicate to check if a value does not equal another value.
    Usage: not_equals(objectToCompareTo)(myObject)
    Usage with Opt: Opt(myObject).filter(not_equals(objectToCompareTo))

    Args:
        obj (Any): The object to compare to.

    Returns:
        Predicate[Any]: The predicate.
    """
    # Reuses equals and not_
    return not_(equals(obj))


def is_blank(obj: Any) -> bool:
    """
    Checks if a value is blank. Returns True in the following conditions:
    - obj is None
    - obj is of type Sized and its len is 0

    Args:
        obj (Any): The object

    Returns:
        bool: True if is blank, False otherwise
    """
    if obj is None:
        return True
    # isinstance check is necessary before len()
    if isinstance(obj, Sized):
        return len(obj) == 0
    # If not None and not Sized, it's not considered blank
    return False


def is_not_blank(obj: Any) -> bool:
    """
    Checks if a value is not blank. Returns True in the following conditions:
    - obj is of type Sized and its len is greater than 0
    - if not of type Sized, object is not None

    Args:
        obj (Any): The object

    Returns:
        bool: True if is not blank, False otherwise
    """
    # Reuses is_blank and not_
    return not_(is_blank)(obj)


def default(default_val: T) -> Callable[[Optional[T]], T]:
    """
    Default value provider. Returns the default value if the input is None.
    Usage: default(defaultValue)(myValue)
    Usage with Opt: Opt(myValue).map(default(defaultValue))

    Args:
        default_val (T): The default value.

    Returns:
        Callable[[Optional[T]], T]: A function that returns the input or the default.
    """

    def wrap(val: Optional[T]) -> T:
        return default_val if val is None else val

    return wrap


def contains(value: Any) -> Predicate[Optional[Union[str, Iterable[Any]]]]:
    """
    Checks if the given value is contained in the call parameter (string or iterable).
    Usage:
    contains("test")("This is the test string") # Returns True
    contains(1)([1, 2, 3]) # Returns True
    Usage with Opt and Stream:
    Opt("This is a test string").filter(contains("test")).get() # Returns True
    Stream([1, 2, 3]).filter(contains(1)).to_list() # Results in [1]

    Args:
        value (Any): The value to check for containment.

    Returns:
        Predicate[Optional[Union[str, Iterable[Any]]]]: A predicate.
    """

    def wrap(val: Optional[Union[str, Iterable[Any]]]) -> bool:
        # Check for None before using 'in'
        return val is not None and value in val

    return Predicate.of(wrap)


def str_contains(value: str) -> Predicate[Optional[str]]:
    """
    Checks if the given string value is contained in the call parameter string.
    Usage: str_contains("test")("This is the test string") # Returns True
    Usage with Opt: Opt("test string").filter(str_contains("test"))

    Args:
        value (str): The substring to check for.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """
    # Correctly uses casting for type specificity
    return cast(Predicate[Optional[str]], contains(value))


def str_contains_ignore_case(value: str) -> Predicate[Optional[str]]:
    """
    Case-insensitive version of str_contains.

    Args:
        value (str): The substring to check for.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # .lower() is the standard way, creates temporary strings.
    def wrap(val: Optional[str]) -> bool:
        return val is not None and value.lower() in val.lower()

    return Predicate.of(wrap)


def str_starts_with(value: str) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string starts with the given value.
    Usage: str_starts_with("test")("test string") # Returns True

    Args:
        value (str): The prefix string.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # Uses efficient built-in str.startswith
    def wrap(val: Optional[str]) -> bool:
        return val is not None and val.startswith(value)

    return Predicate.of(wrap)


def str_starts_with_ignore_case(value: str) -> Predicate[Optional[str]]:
    """
    Case-insensitive version of str_starts_with.

    Args:
        value (str): The prefix string.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # .lower() is standard.
    def wrap(val: Optional[str]) -> bool:
        return val is not None and val.lower().startswith(value.lower())

    return Predicate.of(wrap)


def str_ends_with(value: str) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string ends with the given value.
    Usage: str_ends_with("string")("test string") # Returns True

    Args:
        value (str): The suffix string.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # Uses efficient built-in str.endswith
    def wrap(val: Optional[str]) -> bool:
        return val is not None and val.endswith(value)

    return Predicate.of(wrap)


def str_ends_with_ignore_case(value: str) -> Predicate[Optional[str]]:
    """
    Case-insensitive version of str_ends_with.

    Args:
        value (str): The suffix string.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # .lower() is standard.
    def wrap(val: Optional[str]) -> bool:
        return val is not None and val.lower().endswith(value.lower())

    return Predicate.of(wrap)


def str_matches(pattern: str) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string matches the given regex pattern *at the beginning*.
    Uses `re.match`.
    Usage: str_matches(r"\\d+")("123 abc") # Returns True
    Usage: str_matches(r"\\d+")("abc 123") # Returns False

    Args:
        pattern (str): The regular expression pattern.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # Compile regex for potential minor optimization if used repeatedly,
    # but re.match caches patterns anyway. Keeping it simple is fine.
    # compiled_pattern = re.compile(pattern) # Alternative
    def wrap(val: Optional[str]) -> bool:
        if val is None:
            return False
        # match = compiled_pattern.match(val) # Alternative
        match = re.match(pattern, val)
        return match is not None

    return Predicate.of(wrap)


def str_not_matches(pattern: str) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string does *not* match the given regex pattern *at the beginning*.
    Uses `re.match`.

    Args:
        pattern (str): The regular expression pattern.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """
    # Reuses str_matches and not_
    return not_(str_matches(pattern))


def str_longer_than(length: int) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string's length is greater than the specified value.

    Args:
        length (int): The minimum exclusive length.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    def wrap(val: Optional[str]) -> bool:
        return val is not None and len(val) > length

    return Predicate.of(wrap)


def str_shorter_than(length: int) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string's length is less than the specified value.

    Args:
        length (int): The maximum exclusive length.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    def wrap(val: Optional[str]) -> bool:
        return val is not None and len(val) < length

    return Predicate.of(wrap)


def str_longer_than_or_eq(length: int) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string's length is greater than or equal to the specified value.

    Args:
        length (int): The minimum inclusive length.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    def wrap(val: Optional[str]) -> bool:
        return val is not None and len(val) >= length

    return Predicate.of(wrap)


def str_shorter_than_or_eq(length: int) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string's length is less than or equal to the specified value.

    Args:
        length (int): The maximum inclusive length.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    def wrap(val: Optional[str]) -> bool:
        return val is not None and len(val) <= length

    return Predicate.of(wrap)


def equals_ignore_case(value: str) -> Predicate[Optional[str]]:
    """
    Checks if the call parameter string equals the given value, ignoring case.

    Args:
        value (str): The string to compare against.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # .lower() is standard.
    def wrap(val: Optional[str]) -> bool:
        # Check both for None before comparing
        return val is not None and value.lower() == val.lower()

    return Predicate.of(wrap)


# --- Numeric Predicates ---


def is_even(integer: Optional[int]) -> bool:
    """Checks if an integer is even."""
    return integer is not None and integer % 2 == 0


def is_odd(integer: Optional[int]) -> bool:
    """Checks if an integer is odd."""
    return (
        integer is not None and integer % 2 != 0
    )  # Changed from == 1 for robustness with negative numbers


def is_positive(number: Optional[float]) -> bool:
    """Checks if a number is positive (> 0)."""
    return number is not None and number > 0


def is_negative(number: Optional[float]) -> bool:
    """Checks if a number is negative (< 0)."""
    return number is not None and number < 0


def is_zero(number: Optional[float]) -> bool:
    """Checks if a number is zero (== 0)."""
    # Consider floating point precision issues if exact zero is critical
    return number is not None and number == 0


def is_int(number: Optional[float]) -> bool:
    """Checks if a float represents a whole number."""
    # Handles None check and potential float precision
    return number is not None and number == int(number)


def _beween(interval_start: float, interval_end: float) -> Predicate[Optional[float]]:
    """Checks if a number is strictly between start and end (start < number < end)."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and interval_start < val < interval_end

    return Predicate.of(wrap)


def _beween_closed(
    interval_start: float, interval_end: float
) -> Predicate[Optional[float]]:
    """Checks if a number is between start and end inclusive (start <= number <= end)."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and interval_start <= val <= interval_end

    return Predicate.of(wrap)


def is_beween_closed(
    interval_start: float, interval_end: float
) -> Predicate[Optional[float]]:
    """Checks if a number is between start and end inclusive (start <= number <= end)."""
    return _beween_closed(interval_start, interval_end)


def is_in_interval(
    interval_start: float, interval_end: float
) -> Predicate[Optional[float]]:
    """Checks if a number is between start and end inclusive (start <= number <= end)."""
    return _beween_closed(interval_start, interval_end)


def is_beween(interval_start: float, interval_end: float) -> Predicate[Optional[float]]:
    """Checks if a number is strictly between start and end (start < number < end)."""
    return _beween(interval_start, interval_end)


def is_in_open_interval(
    interval_start: float, interval_end: float
) -> Predicate[Optional[float]]:
    """Checks if a number is strictly between start and end (start < number < end)."""
    return _beween(interval_start, interval_end)


def is_beween_closed_start(
    interval_start: float, interval_end: float
) -> Predicate[Optional[float]]:
    """Checks if a number is between start (inclusive) and end (exclusive) (start <= number < end)."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and interval_start <= val < interval_end

    return Predicate.of(wrap)


def is_beween_closed_end(
    interval_start: float, interval_end: float
) -> Predicate[Optional[float]]:
    """Checks if a number is between start (exclusive) and end (inclusive) (start < number <= end)."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and interval_start < val <= interval_end

    return Predicate.of(wrap)


def is_higher_than(value: float) -> Predicate[Optional[float]]:
    """Checks if a number is strictly greater than the specified value."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and val > value

    return Predicate.of(wrap)


def is_higher_than_or_eq(value: float) -> Predicate[Optional[float]]:
    """Checks if a number is greater than or equal to the specified value."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and val >= value

    return Predicate.of(wrap)


def is_less_than(value: float) -> Predicate[Optional[float]]:
    """Checks if a number is strictly less than the specified value."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and val < value

    return Predicate.of(wrap)


def is_less_than_or_eq(value: float) -> Predicate[Optional[float]]:
    """Checks if a number is less than or equal to the specified value."""

    def wrap(val: Optional[float]) -> bool:
        return val is not None and val <= value

    return Predicate.of(wrap)


# --- Higher-Order Predicates ---


def not_(
    predicate: Union[Predicate[Optional[T]], Callable[[Optional[T]], bool]],
) -> Predicate[Optional[T]]:
    """
    Negates the result of the given predicate. Handles Optional input.
    Usage: not_(is_blank)("test") # Returns True

    Args:
        predicate: The predicate to negate.

    Returns:
        Predicate[Optional[T]]: The negated predicate.
    """
    # Uses predicate_of to handle both Predicate objects and raw callables
    wrapped_predicate = predicate_of(predicate)

    def wrap(val: Optional[T]) -> bool:
        return not wrapped_predicate.apply(val)

    return Predicate.of(wrap)


def not_strict(
    predicate: Callable[[T], bool],
) -> Predicate[T]:
    """
    Negates the result of the given predicate. Requires non-Optional input.
    Useful for maintaining stricter type checking in pipelines.

    Args:
        predicate: The predicate to negate (must accept non-Optional T).

    Returns:
        Predicate[T]: The negated predicate.
    """
    wrapped_predicate = predicate_of(predicate)

    def wrap(val: T) -> bool:
        return not wrapped_predicate.apply(val)

    return Predicate.of(wrap)


# --- Mapping Predicates ---


def has_key(key: Any) -> Predicate[Optional[Mapping[Any, Any]]]:
    """
    Produces a predicate that checks if the argument mapping contains the given key.

    Args:
        key: The key to check for.

    Returns:
        Predicate[Optional[Mapping[Any, Any]]]: The resulting predicate.
    """

    # Using 'key in mapping' is generally preferred over 'key in mapping.keys()'
    def wrap(dct: Optional[Mapping[Any, Any]]) -> bool:
        return dct is not None and key in dct

    return Predicate.of(wrap)


def has_value(value: Any) -> Predicate[Optional[Mapping[Any, Any]]]:
    """
    Produces a predicate that checks if the argument mapping contains the given value.
    Note: This requires iterating through values (O(n)).

    Args:
        value: The value to check for.

    Returns:
        Predicate[Optional[Mapping[Any, Any]]]: The resulting predicate.
    """

    # 'value in mapping.values()' is the standard way
    def wrap(dct: Optional[Mapping[Any, Any]]) -> bool:
        return dct is not None and value in dct.values()

    return Predicate.of(wrap)


def is_key_in(mapping: Mapping[Any, Any]) -> Predicate[Any]:
    """
    Produces a predicate that checks if the argument key is present in the given mapping.

    Args:
        mapping: The mapping to check within.

    Returns:
        Predicate[Any]: The resulting predicate.
    """

    # Using 'key in mapping' is generally preferred
    def wrap(key: Any) -> bool:
        # Check key is not None if that's a requirement, depends on use case
        return key in mapping

    return Predicate.of(wrap)


def is_value_in(mapping: Mapping[Any, Any]) -> Predicate[Any]:
    """
    Produces a predicate that checks if the argument value is present in the given mapping.
    Note: This requires iterating through values (O(n)).

    Args:
        mapping: The mapping to check within.

    Returns:
        Predicate[Any]: The resulting predicate.
    """

    def wrap(value: Any) -> bool:
        return value in mapping.values()

    return Predicate.of(wrap)


def is_truthy(val: Any) -> bool:
    """Checks if a value is considered True in a boolean context."""
    return bool(val)


def is_falsy(val: Any) -> bool:
    """Checks if a value is considered False in a boolean context."""
    return not bool(val)


def is_identity(other: Any) -> Predicate[Any]:
    """Checks if a value is the same object as 'other' (using 'is')."""

    def wrap(val: Any) -> bool:
        return val is other

    return Predicate.of(wrap)


def has_length(length: int) -> Predicate[Optional[Sized]]:
    """Checks if a Sized object has the specified length."""

    def wrap(val: Optional[Sized]) -> bool:
        # is_blank already checks for None
        return not is_blank(val) and len(val) == length  # type: ignore

    return Predicate.of(wrap)


def is_instance(cls: type) -> Predicate[Any]:
    """Checks if a value is an instance of the given class."""

    def wrap(val: Any) -> bool:
        return isinstance(val, cls)

    return Predicate.of(wrap)


def str_fullmatch(pattern: str) -> Predicate[Optional[str]]:
    """
    Checks if the *entire* string matches the given regex pattern (uses re.fullmatch).
    Complements `str_matches` which only checks from the beginning (`re.match`).

    Args:
        pattern (str): The regular expression pattern.

    Returns:
        Predicate[Optional[str]]: A predicate.
    """

    # Consider compiling if performance is critical, but re caches patterns.
    # compiled_pattern = re.compile(pattern)
    def wrap(val: Optional[str]) -> bool:
        if val is None:
            return False
        # match = compiled_pattern.fullmatch(val)
        match = re.fullmatch(pattern, val)
        return match is not None

    return Predicate.of(wrap)


def str_is_alpha(val: Optional[str]) -> bool:
    """Checks if a string is not None and all characters are alphabetic."""
    return val is not None and val.isalpha()


def str_is_digit(val: Optional[str]) -> bool:
    """Checks if a string is not None and all characters are digits."""
    return val is not None and val.isdigit()


def str_is_alnum(val: Optional[str]) -> bool:
    """Checks if a string is not None and all characters are alphanumeric."""
    return val is not None and val.isalnum()


def str_is_lower(val: Optional[str]) -> bool:
    """Checks if a string is not None and all cased characters are lowercase."""
    return val is not None and val.islower()


def str_is_upper(val: Optional[str]) -> bool:
    """Checks if a string is not None and all cased characters are uppercase."""
    return val is not None and val.isupper()


def str_is_space(val: Optional[str]) -> bool:
    """Checks if a string is not None and all characters are whitespace."""
    return val is not None and val.isspace()


def str_is_title(val: Optional[str]) -> bool:
    """Checks if a string is not None and is titlecased."""
    return val is not None and val.istitle()



from jstreams.stream import Stream, Opt, stream, optional, pair_stream, nullable

from jstreams.class_operations import ClassOps

from jstreams.stream_operations import (
    not_null_elements,
    extract_list,
    extract_non_null_list,
    extract_list_strict,
    extract_non_null_list_strict,
)

from jstreams.iterable_operations import (
    find_first,
    take_while,
    drop_while,
    matching,
    reduce,
    take_until,
    drop_until,
    find_last,
)

from jstreams.mapper import (
    map_it,
    flat_map,
    mapper_of,
    mapper_with_of,
    Mapper,
    MapperWith,
)

from jstreams.reducer import reducer_of, Reducer

from jstreams.stream_factories import triplet_stream
from jstreams.stream_predicates import any_of, all_of, none_of, all_none, all_not_none

from jstreams.try_opt import (
    Try,
    ErrorLog,
    try_,
    try_of,
    catch,
    catch_with,
    try_with_resource,
    raises,
)

from jstreams.rx import (
    ObservableSubscription,
    Observable,
    Flowable,
    Single,
    BehaviorSubject,
    PublishSubject,
    ReplaySubject,
    CompletedHandler,
    ErrorHandler,
    DisposeHandler,
    NextHandler,
    Pipe,
    Reduce,
    Filter,
    Map,
    RxOperator,
    BaseFilteringOperator,
    BaseMappingOperator,
    rx_distinct_until_changed,
    rx_reduce,
    rx_filter,
    rx_map,
    rx_take,
    rx_take_while,
    rx_take_until,
    rx_drop_while,
    rx_drop_until,
    rx_drop,
    rx_buffer,
    rx_buffer_count,
    rx_throttle,
    TakeWhile,
    TakeUntil,
    Drop,
    DropUntil,
    DropWhile,
    Take,
    Buffer,
    BufferCount,
    Throttle,
    DistinctUntilChanged,
    RX,
    Timestamped,
    rx_defer,
    rx_distinct,
    rx_element_at,
    rx_empty,
    rx_ignore_all,
    rx_map_to,
    rx_never,
    rx_of_type,
    rx_range,
    rx_scan,
    rx_tap,
    rx_throw,
    rx_timestamp,
    Distinct,
    ElementAt,
    IgnoreAll,
    MapTo,
    Scan,
    Tap,
    Ignore,
    TimestampOperator,
    rx_debounce,
    Debounce,
    rx_combine_latest,
    rx_zip,
)

from jstreams.thread import (
    LoopingThread,
    CallbackLoopingThread,
    cancel_thread,
    Cancellable,
)

from jstreams.timer import (
    Timer,
    Interval,
    CountdownTimer,
    set_timer,
    set_interval,
    clear,
)

from jstreams.ioc import (
    injector,
    AutoInit,
    AutoStart,
    inject,
    inject_optional,
    var,
    InjectedDependency,
    OptionalInjectedDependency,
    resolve,
    resolve_dependencies,
    resolve_variables,
    resolve_all,
    component,
    service,
    Variable,
    StrVariable,
    IntVariable,
    FloatVariable,
    DictVariable,
    SetVariable,
    ListVariable,
    Dependency,
    InjectedVariable,
    inject_args,
    autowired,
    autowired_optional,
    provide,
    provide_variable,
    configuration,
    return_wired,
    return_wired_optional,
)

from jstreams.noop import (
    NoOpCls,
    noop,
)

from jstreams.utils import (
    require_non_null,
    is_number,
    to_int,
    to_float,
    as_list,
    keys_as_list,
    is_mth_or_fn,
    is_not_none,
    is_empty_or_none,
    each,
    dict_update,
    sort,
    Value,
    SynchronizedValue,
    extract,
    identity,
    chunk,
    flatten,
    flatten_deep,
    uniq,
    key_by,
    pick,
    omit,
    head,
    tail,
    last,
    to_nullable,
    repeat_value,
)

from jstreams.predicate import (
    Predicate,
    PredicateWith,
    is_true,
    is_false,
    is_none,
    is_in,
    is_not_in,
    equals,
    is_blank,
    default,
    str_contains,
    str_contains_ignore_case,
    str_starts_with,
    str_starts_with_ignore_case,
    str_ends_with,
    str_ends_with_ignore_case,
    str_matches,
    str_not_matches,
    str_longer_than,
    str_shorter_than,
    str_longer_than_or_eq,
    str_shorter_than_or_eq,
    equals_ignore_case,
    is_even,
    is_odd,
    is_positive,
    is_negative,
    is_zero,
    is_int,
    is_beween,
    is_beween_closed,
    is_beween_closed_start,
    is_beween_closed_end,
    not_,
    not_strict,
    not_equals,
    is_not_blank,
    is_higher_than,
    is_higher_than_or_eq,
    is_less_than,
    is_less_than_or_eq,
    has_key,
    has_value,
    is_in_interval,
    is_in_open_interval,
    is_key_in,
    is_value_in,
    contains,
    is_truthy,
    is_falsy,
    is_identity,
    has_length,
    is_instance,
    str_fullmatch,
    str_is_alnum,
    str_is_alpha,
    str_is_digit,
    str_is_lower,
    str_is_space,
    str_is_title,
    str_is_upper,
    predicate_of,
    predicate_with_of,
)

from jstreams.match import (
    Case,
    Match,
    DefaultCase,
    case,
    match,
    match_opt,
    default_case,
)

from jstreams.tuples import (
    Pair,
    pair,
    Triplet,
    triplet,
    left_matches,
    right_matches,
    middle_matches,
    pair_of,
    triplet_of,
)

from jstreams.collectors import (
    grouping_by,
    joining,
    grouping_by_mapping,
    Collectors,
)

from jstreams.state import (
    default_state,
    use_state,
    use_async_state,
    null_state,
)

from jstreams.scheduler import (
    scheduler,
    schedule_daily,
    schedule_duration,
    schedule_hourly,
    schedule_periodic,
    Duration,
)

from jstreams.eventing import (
    events,
    event,
    on_event,
    managed_events,
    dispose_managed_events_from,
    wait_for_event,  # Add wait_for_event to exports
)

from jstreams.annotations import (
    builder,
    getter,
    setter,
    locked,
    synchronized_static,
    synchronized,
    all_args,
    required_args,
    validate_args,
    default_on_error,
)

from jstreams.serialize import (
    json_serializable,
    json_deserialize,
    json_serialize,
    json_standard_serializable,
    json_serialize_return,
    json_deserialize_list,
    json_serialize_return_list,
)

__all__ = [
    "each",
    "dict_update",
    "Stream",
    "find_first",
    "map_it",
    "matching",
    "flat_map",
    "reduce",
    "take_while",
    "drop_while",
    "is_not_none",
    "sort",
    "Opt",
    "ClassOps",
    "stream",
    "optional",
    "Try",
    "ErrorLog",
    "ObservableSubscription",
    "Observable",
    "Flowable",
    "Single",
    "BehaviorSubject",
    "PublishSubject",
    "ReplaySubject",
    "Pipe",
    "Reduce",
    "Filter",
    "Map",
    "rx_reduce",
    "rx_filter",
    "rx_map",
    "RxOperator",
    "BaseFilteringOperator",
    "BaseMappingOperator",
    "LoopingThread",
    "CallbackLoopingThread",
    "Timer",
    "Interval",
    "CountdownTimer",
    "cancel_thread",
    "set_timer",
    "set_interval",
    "clear",
    "injector",
    "NoOpCls",
    "noop",
    "inject",
    "inject_optional",
    "var",
    "require_non_null",
    "is_number",
    "to_int",
    "to_float",
    "as_list",
    "keys_as_list",
    "is_true",
    "is_false",
    "is_none",
    "is_in",
    "is_not_in",
    "equals",
    "is_blank",
    "default",
    "all_none",
    "all_not_none",
    "str_contains",
    "str_contains_ignore_case",
    "str_starts_with",
    "str_starts_with_ignore_case",
    "str_ends_with",
    "str_ends_with_ignore_case",
    "str_matches",
    "str_not_matches",
    "str_longer_than",
    "str_shorter_than",
    "str_longer_than_or_eq",
    "str_shorter_than_or_eq",
    "equals_ignore_case",
    "is_even",
    "is_odd",
    "is_positive",
    "is_negative",
    "is_zero",
    "is_int",
    "is_beween",
    "is_beween_closed",
    "is_beween_closed_start",
    "is_beween_closed_end",
    "not_",
    "not_strict",
    "not_equals",
    "is_not_blank",
    "AutoStart",
    "AutoInit",
    "is_mth_or_fn",
    "Case",
    "Match",
    "DefaultCase",
    "case",
    "match",
    "match_opt",
    "default_case",
    "is_higher_than",
    "is_higher_than_or_eq",
    "is_less_than",
    "is_less_than_or_eq",
    "any_of",
    "all_of",
    "none_of",
    "predicate_of",
    "predicate_with_of",
    "mapper_of",
    "mapper_with_of",
    "Predicate",
    "Mapper",
    "PredicateWith",
    "MapperWith",
    "Reducer",
    "reducer_of",
    "has_key",
    "has_value",
    "is_in_interval",
    "is_in_open_interval",
    "contains",
    "rx_take",
    "rx_take_while",
    "rx_take_until",
    "rx_drop_while",
    "rx_drop_until",
    "rx_drop",
    "Cancellable",
    "is_empty_or_none",
    "CompletedHandler",
    "ErrorHandler",
    "DisposeHandler",
    "NextHandler",
    "TakeWhile",
    "TakeUntil",
    "Drop",
    "DropUntil",
    "DropWhile",
    "Take",
    "Pair",
    "pair",
    "Triplet",
    "triplet",
    "left_matches",
    "right_matches",
    "middle_matches",
    "pair_of",
    "triplet_of",
    "is_key_in",
    "is_value_in",
    "InjectedDependency",
    "OptionalInjectedDependency",
    "resolve_dependencies",
    "resolve_variables",
    "Variable",
    "StrVariable",
    "IntVariable",
    "FloatVariable",
    "Dependency",
    "InjectedVariable",
    "component",
    "inject_args",
    "grouping_by",
    "joining",
    "Collectors",
    "DictVariable",
    "default_state",
    "use_state",
    "use_async_state",
    "null_state",
    "extract_list",
    "extract_non_null_list",
    "not_null_elements",
    "ListVariable",
    "SetVariable",
    "service",
    "autowired",
    "autowired_optional",
    "provide",
    "provide_variable",
    "configuration",
    "return_wired",
    "return_wired_optional",
    "try_",
    "try_of",
    "scheduler",
    "schedule_daily",
    "schedule_duration",
    "schedule_hourly",
    "schedule_periodic",
    "Duration",
    "pair_stream",
    "triplet_stream",
    "catch",
    "catch_with",
    "try_with_resource",
    "resolve",
    "events",
    "event",
    "resolve_all",
    "Value",
    "SynchronizedValue",
    "extract",
    "identity",
    "builder",
    "getter",
    "setter",
    "locked",
    "synchronized_static",
    "synchronized",
    "all_args",
    "required_args",
    "validate_args",
    "default_on_error",
    "is_truthy",
    "is_falsy",
    "is_identity",
    "has_length",
    "is_instance",
    "str_fullmatch",
    "str_is_alnum",
    "str_is_alpha",
    "str_is_digit",
    "str_is_lower",
    "str_is_space",
    "str_is_title",
    "str_is_upper",
    "grouping_by_mapping",
    "RX",
    "take_until",
    "drop_until",
    "find_last",
    "on_event",
    "managed_events",
    "dispose_managed_events_from",
    "wait_for_event",
    "nullable",
    "rx_distinct_until_changed",
    "rx_buffer",
    "rx_buffer_count",
    "rx_throttle",
    "Buffer",
    "BufferCount",
    "Throttle",
    "DistinctUntilChanged",
    "to_nullable",
    "raises",
    "json_serializable",
    "json_deserialize",
    "json_serialize",
    "json_standard_serializable",
    "json_serialize_return",
    "json_serialize_return_list",
    "json_deserialize_list",
    "json_serialize_return_list",
    "chunk",
    "flatten",
    "flatten_deep",
    "uniq",
    "key_by",
    "pick",
    "omit",
    "head",
    "tail",
    "last",
    "extract_list_strict",
    "extract_non_null_list_strict",
    "repeat_value",
    "Timestamped",
    "rx_defer",
    "rx_distinct",
    "rx_element_at",
    "rx_empty",
    "rx_ignore_all",
    "rx_map_to",
    "rx_never",
    "rx_of_type",
    "rx_range",
    "rx_scan",
    "rx_tap",
    "rx_throw",
    "rx_timestamp",
    "Distinct",
    "ElementAt",
    "IgnoreAll",
    "MapTo",
    "Scan",
    "Tap",
    "Ignore",
    "TimestampOperator",
    "rx_debounce",
    "Debounce",
    "rx_combine_latest",
    "rx_zip",
]
